from collections.abc import Sequence
from time import time
from typing import cast

from numpy import array, exp, ndarray
from pygfunction.boreholes import Borehole

from ghedesigner.constants import DEG_TO_RAD, MONTHS_IN_YEAR
from ghedesigner.enums import DesignGeomType, FlowConfigType, PipeType, TimestepType
from ghedesigner.ghe.boreholes.single_u_borehole import SingleUTube
from ghedesigner.ghe.coordinates import rectangle
from ghedesigner.ghe.design.base import DesignBase
from ghedesigner.ghe.design.birectangle import DesignBiRectangle, GeometricConstraintsBiRectangle
from ghedesigner.ghe.design.birectangle_constrained import (
    DesignBiRectangleConstrained,
    GeometricConstraintsBiRectangleConstrained,
)
from ghedesigner.ghe.design.bizoned import DesignBiZoned, GeometricConstraintsBiZoned
from ghedesigner.ghe.design.near_square import DesignNearSquare, GeometricConstraintsNearSquare
from ghedesigner.ghe.design.rectangle import DesignRectangle, GeometricConstraintsRectangle
from ghedesigner.ghe.design.rowwise import DesignRowWise, GeometricConstraintsRowWise
from ghedesigner.ghe.gfunction import calculate_g_function
from ghedesigner.ghe.ground_heat_exchangers import GHE
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.media import Fluid, Grout, Soil
from ghedesigner.utilities import combine_sts_lts, eskilson_log_times, get_loads


class GroundHeatExchanger:  # TODO: Rename this.  Just GHEDesignerManager?  GHEDesigner?
    def __init__(
        self,
        grout_conductivity: float,
        grout_rho_cp: float,
        soil_conductivity: float,
        soil_rho_cp: float,
        soil_undisturbed_temperature: float,
        borehole_buried_depth: float,
        borehole_radius: float,
        pipe_arrangement_type: PipeType,
        pipe_parameters: dict,
        fluid_name: str = "Water",
        fluid_concentration_percent: float = 0.0,
        fluid_temperature: float = 20.0,
    ) -> None:
        self.fluid = Fluid(fluid_name, fluid_temperature, fluid_concentration_percent)
        self.grout = Grout(grout_conductivity, grout_rho_cp)
        self.soil = Soil(soil_conductivity, soil_rho_cp, soil_undisturbed_temperature)
        if pipe_arrangement_type == PipeType.SINGLEUTUBE:
            params = ["conductivity", "rho_cp", "inner_diameter", "outer_diameter", "shank_spacing", "roughness"]
            if not all(x in pipe_parameters for x in params):
                raise ValueError(f"pipe_arrangement_type of {pipe_arrangement_type!s} requires these inputs: {params}")
            pipe_parameters["num_pipes"] = 1
            self.pipe = Pipe.init_single_u_tube(**pipe_parameters)
        elif pipe_arrangement_type == PipeType.DOUBLEUTUBESERIES:
            params = ["conductivity", "rho_cp", "inner_diameter", "outer_diameter", "shank_spacing", "roughness"]
            if not all(x in pipe_parameters for x in params):
                raise ValueError(f"pipe_arrangement_type of {pipe_arrangement_type!s} requires these inputs: {params}")
            self.pipe = Pipe.init_double_u_tube_series(**pipe_parameters)
        elif pipe_arrangement_type == PipeType.DOUBLEUTUBEPARALLEL:
            params = ["conductivity", "rho_cp", "inner_diameter", "outer_diameter", "shank_spacing", "roughness"]
            if not all(x in pipe_parameters for x in params):
                raise ValueError(f"pipe_arrangement_type of {pipe_arrangement_type!s} requires these inputs: {params}")
            self.pipe = Pipe.init_double_u_tube_parallel(**pipe_parameters)
        else:  # Assuming coaxial
            params = [
                "conductivity_inner",
                "rho_cp",
                "conductivity_outer",
                "inner_pipe_d_in",
                "inner_pipe_d_out",
                "outer_pipe_d_in",
                "outer_pipe_d_out",
            ]
            if not all(x in pipe_parameters for x in params):
                raise ValueError(f"pipe_arrangement_type of {PipeType.COAXIAL!s} requires these inputs: {params}")
            pipe_parameters["conductivity"] = (
                pipe_parameters["conductivity_inner"],
                pipe_parameters["conductivity_outer"],
            )
            del pipe_parameters["conductivity_inner"]
            del pipe_parameters["conductivity_outer"]
            self.pipe = Pipe.init_coaxial(**pipe_parameters)
        self.pygfunction_borehole = Borehole(100, borehole_buried_depth, borehole_radius, x=0.0, y=0.0)

    @classmethod
    def init_from_dictionary(cls, ghe_dict: dict, fluid_inputs: dict | None = None) -> "GroundHeatExchanger":
        """
        Initialize a GroundHeatExchanger object from input dictionaries, performing validation and ultimately calling
        the main object constructor.
        :param ghe_dict: Dictionary of ground heat exchanger parameters, see the input schema specification for required
                         inputs in the ground_heat_exchanger schema field.
        :param fluid_inputs: Optional dictionary of fluid input parameters, see the input schema fluid spec for details.
        :return: GroundHeatExchanger object.
        # TODO: Add validation back in to the input fields
        """
        grout_parameters: dict = ghe_dict["grout"]
        g_c: float = grout_parameters["conductivity"]
        g_rho_cp: float = grout_parameters["rho_cp"]

        soil_parameters: dict = ghe_dict["soil"]
        s_k: float = soil_parameters["conductivity"]
        s_rho_cp: float = soil_parameters["rho_cp"]
        s_temp: float = soil_parameters["undisturbed_temp"]

        borehole_parameters: dict = ghe_dict["borehole"]
        buried_depth: float = borehole_parameters["buried_depth"]
        diameter: float = borehole_parameters["diameter"]
        radius: float = diameter / 2.0

        fluid_dict = (
            fluid_inputs if fluid_inputs else {"fluid_name": "Water", "concentration_percent": 0.0, "temperature": 20.0}
        )
        fluid_name = fluid_dict.get("fluid_name", "Water")
        concentration_percent = fluid_dict.get("concentration_percent", 0.0)
        temperature = fluid_dict.get("temperature", 20.0)

        pipe_parameters: dict = ghe_dict["pipe"]
        pipe_type: PipeType = PipeType(pipe_parameters["arrangement"].upper())
        del pipe_parameters["arrangement"]

        ghe: GroundHeatExchanger = cls(
            g_c,
            g_rho_cp,
            s_k,
            s_rho_cp,
            s_temp,
            buried_depth,
            radius,
            pipe_type,
            pipe_parameters,
            fluid_name,
            concentration_percent,
            temperature,
        )
        return ghe

    def design_and_size_ghe(self, ghe_dict: dict, end_month: int, loads_override: list[float] | None = None):
        ghe_loads = loads_override if loads_override else get_loads(ghe_dict["loads"])

        if (end_month % MONTHS_IN_YEAR) > 0:
            raise ValueError(f"end_month must be a multiple of {MONTHS_IN_YEAR}")

        flow_type_str: str = ghe_dict["flow_type"]
        flow_type = FlowConfigType(flow_type_str.upper())
        flow_rate: float = ghe_dict["flow_rate"]

        # grab some design conditions
        design_parameters = ghe_dict["design"]
        continue_if_design_unmet: bool = design_parameters.get("continue_if_design_unmet", False)
        min_eft: float = design_parameters["min_eft"]
        max_eft: float = design_parameters["max_eft"]
        max_height: float = design_parameters["max_height"]
        min_height: float = design_parameters["min_height"]
        max_boreholes: int | None = design_parameters.get("max_boreholes")
        # check_arg_bounds(min_eft, max_eft, "min_eft", "max_eft")

        # set up the geometry constraints section
        geom = ghe_dict["geometric_constraints"]
        geometry_map = {geom.name: geom for geom in DesignGeomType}
        geom_type = geometry_map.get(geom["method"].upper())
        design: DesignBase

        match geom_type:
            case DesignGeomType.RECTANGLE:
                # max_height: float, min_height: float, length: float, width: float, b_min: float, b_max: float
                rect_geometry: GeometricConstraintsRectangle = GeometricConstraintsRectangle(
                    length=geom["length"],
                    width=geom["width"],
                    b_min=geom["b_min"],
                    b_max=geom["b_max"],
                )
                design = DesignRectangle(
                    flow_rate,
                    self.pygfunction_borehole,
                    self.fluid,
                    self.pipe,
                    self.grout,
                    self.soil,
                    1,
                    end_month,
                    max_eft,
                    min_eft,
                    max_height,
                    min_height,
                    continue_if_design_unmet,
                    max_boreholes,
                    rect_geometry,
                    ghe_loads,
                    flow_type=flow_type,
                    method=TimestepType.HYBRID,
                )
            case DesignGeomType.NEARSQUARE:
                near_sq_geometry: GeometricConstraintsNearSquare = GeometricConstraintsNearSquare(
                    b=geom["b"],
                    length=geom["length"],
                )
                design = DesignNearSquare(
                    flow_rate,
                    self.pygfunction_borehole,
                    self.fluid,
                    self.pipe,
                    self.grout,
                    self.soil,
                    1,
                    end_month,
                    max_eft,
                    min_eft,
                    max_height,
                    min_height,
                    continue_if_design_unmet,
                    max_boreholes,
                    near_sq_geometry,
                    ghe_loads,
                    flow_type=flow_type,
                    method=TimestepType.HYBRID,
                )
            case DesignGeomType.BIRECTANGLE:
                bi_rect_geometry: GeometricConstraintsBiRectangle = GeometricConstraintsBiRectangle(
                    length=geom["length"],
                    width=geom["width"],
                    b_min=geom["b_min"],
                    b_max_x=geom["b_max_x"],
                    b_max_y=geom["b_max_y"],
                )
                design = DesignBiRectangle(
                    flow_rate,
                    self.pygfunction_borehole,
                    self.fluid,
                    self.pipe,
                    self.grout,
                    self.soil,
                    1,
                    end_month,
                    max_eft,
                    min_eft,
                    max_height,
                    min_height,
                    continue_if_design_unmet,
                    max_boreholes,
                    bi_rect_geometry,
                    ghe_loads,
                    flow_type=flow_type,
                    method=TimestepType.HYBRID,
                )
            case DesignGeomType.BIZONEDRECTANGLE:
                bi_zoned_geometry: GeometricConstraintsBiZoned = GeometricConstraintsBiZoned(
                    length=geom["length"],
                    width=geom["width"],
                    b_min=geom["b_min"],
                    b_max_x=geom["b_max_x"],
                    b_max_y=geom["b_max_y"],
                )
                design = DesignBiZoned(
                    flow_rate,
                    self.pygfunction_borehole,
                    self.fluid,
                    self.pipe,
                    self.grout,
                    self.soil,
                    1,
                    end_month,
                    max_eft,
                    min_eft,
                    max_height,
                    min_height,
                    continue_if_design_unmet,
                    max_boreholes,
                    bi_zoned_geometry,
                    ghe_loads,
                    flow_type=flow_type,
                    method=TimestepType.HYBRID,
                )
            case DesignGeomType.BIRECTANGLECONSTRAINED:
                no_go_boundaries = geom.get("no_go_boundaries", None)
                bi_rect_const_geometry: GeometricConstraintsBiRectangleConstrained = (
                    GeometricConstraintsBiRectangleConstrained(
                        b_min=geom["b_min"],
                        b_max_x=geom["b_max_x"],
                        b_max_y=geom["b_max_y"],
                        property_boundary=geom["property_boundary"],
                        no_go_boundaries=no_go_boundaries,
                    )
                )
                design = DesignBiRectangleConstrained(
                    flow_rate,
                    self.pygfunction_borehole,
                    self.fluid,
                    self.pipe,
                    self.grout,
                    self.soil,
                    1,
                    end_month,
                    max_eft,
                    min_eft,
                    max_height,
                    min_height,
                    continue_if_design_unmet,
                    max_boreholes,
                    bi_rect_const_geometry,
                    ghe_loads,
                    flow_type=flow_type,
                    method=TimestepType.HYBRID,
                )
            case DesignGeomType.ROWWISE:
                # use perimeter calculations if present
                perimeter_spacing_ratio = geom.get("perimeter_spacing_ratio", None)
                spacing_step = geom.get("spacing_step", 0)
                no_go_boundaries = geom.get("no_go_boundaries", None)
                geometry_row: GeometricConstraintsRowWise = GeometricConstraintsRowWise(
                    perimeter_spacing_ratio=perimeter_spacing_ratio,
                    max_spacing=geom["max_spacing"],
                    min_spacing=geom["min_spacing"],
                    spacing_step=spacing_step,
                    max_rotation=geom["max_rotation"] * DEG_TO_RAD,
                    min_rotation=geom["min_rotation"] * DEG_TO_RAD,
                    rotate_step=geom["rotate_step"],
                    property_boundary=geom["property_boundary"],
                    no_go_boundaries=no_go_boundaries,
                )
                design = DesignRowWise(
                    flow_rate,
                    self.pygfunction_borehole,
                    self.fluid,
                    self.pipe,
                    self.grout,
                    self.soil,
                    1,
                    end_month,
                    max_eft,
                    min_eft,
                    max_height,
                    min_height,
                    continue_if_design_unmet,
                    max_boreholes,
                    geometry_row,
                    ghe_loads,
                    flow_type=flow_type,
                    method=TimestepType.HYBRID,
                )
            case _:
                raise ValueError(f'DesignGeomType "{geom_type}" not supported')

        start_time = time()
        search = design.find_design()  # TODO: I wonder if it would simplify things to just return the GHE object
        search_time = time() - start_time
        found_ghe = cast(GHE, search.ghe)
        found_ghe.compute_g_functions(min_height, max_height)
        found_ghe.size(TimestepType.HYBRID, max_height, min_height, max_eft, min_eft)
        return search, search_time, found_ghe

    def get_g_function(self, ghe_dict: dict, boundary_condition="MIFT") -> tuple[ndarray, ndarray, ndarray]:
        # TODO: Create a SingleUTube class or something in order to get the STS stitched up
        pre_designed = ghe_dict["pre_designed"]
        borehole_height: float = pre_designed["H"]
        if pre_designed["arrangement"] == "MANUAL":
            x_positions: Sequence[float] = pre_designed["x"]
            y_positions: Sequence[float] = pre_designed["y"]
            if len(x_positions) != len(y_positions):
                raise RuntimeError("Borehole location coordinate mismatch, make sure length of x and y are equal")
            locations = list(zip(x_positions, y_positions))
        elif pre_designed["arrangement"] == "RECTANGLE":
            num_bh_x = pre_designed["boreholes_in_x_dimension"]
            num_bh_y = pre_designed["boreholes_in_y_dimension"]
            spacing_x = pre_designed["spacing_in_x_dimension"]
            spacing_y = pre_designed["spacing_in_y_dimension"]
            locations = rectangle(num_bh_x, num_bh_y, spacing_x, spacing_y)
        else:
            raise RuntimeError("Invalid arrangement type for pre_designed borehole field")

        nbh = len(locations)
        flow_rate: float = ghe_dict["flow_rate"]
        flow_type_str: str = str(ghe_dict["flow_type"]).upper()

        if flow_type_str == FlowConfigType.BOREHOLE.name:
            m_flow_borehole = flow_rate * self.fluid.rho / 1000  # conv lps to m3s to kgs
        elif flow_type_str == FlowConfigType.SYSTEM.name:
            m_flow_ghe = flow_rate * self.fluid.rho / 1000  # conv lps to m3s to kgs
            m_flow_borehole = m_flow_ghe / nbh
        else:
            raise NotImplementedError(f"FlowConfigType {flow_type_str} not implemented.")

        self.pygfunction_borehole.H = borehole_height
        ts = borehole_height**2 / (9 * self.soil.alpha)
        log_time_lts = eskilson_log_times()
        time_values = exp(log_time_lts) * ts

        g_lts = calculate_g_function(
            m_flow_borehole,
            self.pipe.type,
            time_values,
            locations,
            self.pygfunction_borehole,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            boundary_condition=boundary_condition,
        )

        single_u_bh = SingleUTube(
            m_flow_borehole, self.fluid, self.pygfunction_borehole, self.pipe, self.grout, self.soil
        )

        log_time_sts, g_sts = single_u_bh.calc_sts_g_functions()
        g_bhw = single_u_bh.g_bhw

        g_interp = combine_sts_lts(
            log_time_lts,
            g_lts.tolist(),
            log_time_sts.tolist(),
            g_sts.tolist(),
        )

        g_bhw_interp = combine_sts_lts(
            log_time_lts,
            g_lts.tolist(),
            log_time_sts.tolist(),
            g_bhw.tolist(),
        )

        log_time_to_write = array(log_time_sts.tolist() + log_time_lts)
        g_to_write = g_interp(log_time_to_write)
        g_bhw_to_write = g_bhw_interp(log_time_to_write)

        return log_time_to_write, g_to_write, g_bhw_to_write

    # def write_input_file(self, output_file_path: Path, simulation_parameters: SimulationParameters) -> None:
    #     """
    #     Writes an input file based on current simulation configuration.
    #
    #     :param output_file_path: output directory to write input file.
    #     :raises AttributeError: If necessary class attributes are not set.
    #     :raises ValueError: If the pipe type is not supported.
    #     """
    #     # TODO: geometric constraints are currently held in two places
    #     #       SimulationParameters and GeometricConstraints
    #     #       these should be consolidated
    #     d_geo = self._geometric_constraints.to_input()
    #     d_geo["max_height"] = simulation_parameters.max_height
    #     d_geo["min_height"] = simulation_parameters.min_height
    #
    #     # TODO: data held in different places
    #     d_des = self._design.to_input()
    #     d_des["max_eft"] = simulation_parameters.max_EFT_allowable
    #     d_des["min_eft"] = simulation_parameters.min_EFT_allowable
    #
    #     if simulation_parameters.max_boreholes is not None:
    #         d_des["max_boreholes"] = simulation_parameters.max_boreholes
    #     if simulation_parameters.continue_if_design_unmet is True:
    #         d_des["continue_if_design_unmet"] = simulation_parameters.continue_if_design_unmet
    #
    #     # pipe data
    #     d_pipe = {"rho_cp": self.pipe.rho_cp, "roughness": self.pipe.roughness}
    #
    #     if self.pipe.type in [BHPipeType.SINGLEUTUBE, BHPipeType.DOUBLEUTUBEPARALLEL, BHPipeType.DOUBLEUTUBESERIES]:
    #         d_pipe["inner_diameter"] = self.pipe.r_in * 2.0
    #         d_pipe["outer_diameter"] = self.pipe.r_out * 2.0
    #         d_pipe["shank_spacing"] = self.pipe.s
    #         d_pipe["conductivity"] = self.pipe.k
    #     elif self.pipe.type == BHPipeType.COAXIAL:
    #         d_pipe["inner_pipe_d_in"] = self.pipe.r_in[0] * 2.0
    #         d_pipe["inner_pipe_d_out"] = self.pipe.r_in[1] * 2.0
    #         d_pipe["outer_pipe_d_in"] = self.pipe.r_out[0] * 2.0
    #         d_pipe["outer_pipe_d_out"] = self.pipe.r_out[1] * 2.0
    #         d_pipe["conductivity_inner"] = self.pipe.k[0]
    #         d_pipe["conductivity_outer"] = self.pipe.k[1]
    #     else:
    #         raise ValueError(f"Invalid pipe type '{self.pipe.type.name if self.pipe.type else 'None'}'")
    #
    #     d_pipe["arrangement"] = self.pipe.type.name
    #
    #     d = {
    #         "fluid": self.fluid.to_input(),
    #         "grout": self.grout.to_input(),
    #         "soil": self.soil.to_input(),
    #         "pipe": d_pipe,
    #         # "borehole": self._borehole.to_input(),
    #         # "simulation": self._simulation_parameters.to_input(),
    #         "geometric_constraints": d_geo,
    #         "design": d_des,
    #         "loads": {"ground_loads": self._ground_loads},
    #     }
    #
    #     output_file_path.parent.mkdir(parents=True, exist_ok=True)
    #     output_file_path.write_text(dumps(d, sort_keys=True, indent=2, separators=(",", ": ")))



from scp.ethyl_alcohol import EthylAlcohol
from scp.ethylene_glycol import EthyleneGlycol
from scp.methyl_alcohol import MethylAlcohol
from scp.propylene_glycol import PropyleneGlycol
from scp.water import Water

from ghedesigner.enums import FluidType


class Fluid:
    def __init__(self, fluid_name: str, temperature: float = 20, percent: float = 0) -> None:
        self.name = fluid_name
        self.fluid_type = self.get_fluid_type(fluid_name)
        self.temperature = temperature
        self.concentration_percent = percent

        concentration_frac = self.concentration_percent / 100
        if self.fluid_type == FluidType.ETHYLALCOHOL:
            self._fluid = EthylAlcohol(concentration_frac)
        elif self.fluid_type == FluidType.ETHYLENEGLYCOL:
            self._fluid = EthyleneGlycol(concentration_frac)
        elif self.fluid_type == FluidType.METHYLALCOHOL:
            self._fluid = MethylAlcohol(concentration_frac)
        elif self.fluid_type == FluidType.PROPYLENEGLYCOL:
            self._fluid = PropyleneGlycol(concentration_frac)
        elif self.fluid_type == FluidType.WATER:
            self._fluid = Water()

        # supported props
        self.cp: float = 0.0
        self.k: float = 0.0
        self.mu: float = 0.0
        self.rho: float = 0.0
        self.rho_cp: float = 0.0
        self.update_props_with_new_temp(temperature)

    @staticmethod
    def get_fluid_type(fluid_name: str) -> FluidType:
        fluid_name_upper = fluid_name.upper()
        if fluid_name_upper in ["MEA", "ETHYLALCOHOL", "ETHYL ALCOHOL"]:
            return FluidType.ETHYLALCOHOL
        if fluid_name_upper in ["MEG", "ETHYLENEGLYCOL", "ETHYLENE GLYCOL"]:
            return FluidType.ETHYLENEGLYCOL
        if fluid_name_upper in ["MMA", "METHYLALCOHOL", "METHYL ALCOHOL"]:
            return FluidType.METHYLALCOHOL
        if fluid_name_upper in ["MPG", "PROPYLENEGLYCOL", "PROPYLENE GLYCOL"]:
            return FluidType.PROPYLENEGLYCOL
        if fluid_name_upper == "WATER":
            return FluidType.WATER

        raise ValueError(f'Unsupported fluid type "{fluid_name}"')

    def update_props_with_new_temp(self, temperature: float) -> None:
        self.temperature = temperature
        self.cp = self._fluid.cp(self.temperature)
        self.k = self._fluid.k(self.temperature)
        self.mu = self._fluid.mu(self.temperature)
        self.rho = self._fluid.rho(self.temperature)
        self.rho_cp = self.rho * self.cp


class ThermalProperty:
    def __init__(self, k, rho_cp: float) -> None:
        self.k = k  # Thermal conductivity (W/m.K)
        self.rho_cp = rho_cp  # Volumetric heat capacity (J/K.m3)

    def as_dict(self) -> dict:
        output = {
            "type": str(self.__class__),
            "thermal_conductivity": {"value": self.k, "units": "W/m-K"},
            "volumetric_heat_capacity": {"value": self.rho_cp, "units": "J/K-m3"},
        }
        return output

    def to_input(self) -> dict:
        return {"conductivity": self.k, "rho_cp": self.rho_cp}


class Grout(ThermalProperty):
    pass


class Soil(ThermalProperty):
    def __init__(self, k: float, rho_cp: float, ugt: float) -> None:
        # Make variables from ThermalProperty available to Pipe
        super().__init__(k, rho_cp)

        # Soil specific parameters
        self.ugt = ugt
        self.alpha = k / rho_cp

    def as_dict(self) -> dict:
        output = super().as_dict()
        output["undisturbed_ground_temperature"] = {"value": self.ugt, "units": "C"}
        return output

    def to_input(self) -> dict:
        return {"conductivity": self.k, "rho_cp": self.rho_cp, "undisturbed_temp": self.ugt}



from math import floor

from ghedesigner.constants import HRS_IN_DAY


def hours_to_month(hours: float) -> float:
    """
    Convert a number of hours into fractional months, assuming a non-leap year calendar.
    """
    days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    hrs_year = [HRS_IN_DAY * d for d in days]
    total_year = sum(hrs_year)
    years = floor(hours / total_year)
    rem = hours - years * total_year

    # find month index
    total = 0
    month = 0
    for idx, h in enumerate(hrs_year):
        total += h
        if rem <= total:
            month = idx
            break

    prev_total = total - hrs_year[month]
    month_frac = (rem - prev_total) / hrs_year[month]
    return years * 12 + month + month_frac


def ghe_time_convert(hours: int) -> tuple[int, int, int]:
    """
    Convert total hours into (month, day, hour) tuple in a non-leap year.
    """
    days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    hrs_year = [HRS_IN_DAY * d for d in days]

    total = 0
    month = 0
    for idx, h in enumerate(hrs_year):
        if hours < total + h:
            month = idx
            break
        total += h

    hr_into_month = hours - total
    day = floor(hr_into_month / HRS_IN_DAY) + 1
    hour_in_day = (hr_into_month % HRS_IN_DAY) + 1
    return month + 1, day, int(hour_in_day)



import numpy as np
from pygfunction.boreholes import Borehole

from ghedesigner.enums import FlowConfigType, TimestepType
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.ghe.search.bisection_1d import Bisection1D
from ghedesigner.media import Fluid, Grout, Soil


class BisectionZD(Bisection1D):
    def __init__(
        self,
        coordinates_domain_nested: list,
        field_descriptors: list,
        v_flow: float,
        borehole: Borehole,
        fluid: Fluid,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        max_boreholes: int | None,
        min_height: float,
        max_height: float,
        continue_if_design_unmet: bool,
        start_month: int,
        end_month: int,
        min_eft: float,
        max_eft: float,
        hourly_extraction_ground_loads: list,
        method: TimestepType,
        flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
        max_iter=15,
        disp=False,
        field_type="N/A",
        load_years=None,
    ) -> None:
        if load_years is None:
            load_years = [2019]
        if disp:
            print("Note: This design routine currently requires several bisection searches.")

        # Get a coordinates domain for initialization
        coordinates_domain = coordinates_domain_nested[0]
        super().__init__(
            coordinates_domain,
            field_descriptors[0],
            v_flow,
            borehole,
            fluid,
            pipe,
            grout,
            soil,
            max_boreholes,
            min_height,
            max_height,
            continue_if_design_unmet,
            start_month,
            end_month,
            min_eft,
            max_eft,
            hourly_extraction_ground_loads,
            method=method,
            flow_type=flow_type,
            max_iter=max_iter,
            disp=disp,
            search=False,
            field_type=field_type,
            load_years=load_years,
        )
        self.max_height = max_height
        self.coordinates_domain_nested = coordinates_domain_nested
        self.nested_fieldDescriptors = field_descriptors
        self.calculated_temperatures_nested: dict[int, dict[int, np.float64]] = {}
        # Tack on one borehole at the beginning to provide a high excess
        # temperature
        outer_domain = [coordinates_domain_nested[0][0]]
        outer_descriptors = [field_descriptors[0][0]]
        for cdn, fd in zip(coordinates_domain_nested, field_descriptors):
            outer_domain.append(cdn[-1])
            outer_descriptors.append(fd[-1])

        self.coordinates_domain = outer_domain
        self.fieldDescriptors = outer_descriptors

        self.selection_key_outer, _ = self.search()
        if self.selection_key_outer > 0:
            self.selection_key_outer -= 1
        self.calculated_heights: dict[int, float] = {}

        self.selection_key, self.selected_coordinates = self.search_successive()

    def search_successive(self, max_iter=None):
        if max_iter is None:
            max_iter = self.selection_key_outer + 7

        i = self.selection_key_outer

        old_height = 99999

        while i < len(self.coordinates_domain_nested) and i < max_iter:
            self.coordinates_domain = self.coordinates_domain_nested[i]
            self.fieldDescriptors = self.nested_fieldDescriptors[i]
            self.calculated_temperatures = {}
            try:
                selection_key, selected_coordinates = self.search()
            except ValueError:
                break
            self.calculated_temperatures_nested[i] = self.calculated_temperatures

            self.ghe.compute_g_functions(self.min_height, self.max_height)
            self.ghe.size(
                method=TimestepType.HYBRID,
                max_height=self.max_height,
                min_height=self.min_height,
                design_max_eft=self.max_eft,
                design_min_eft=self.min_eft,
            )

            nbh = len(selected_coordinates)
            total_drilling = nbh * self.ghe.bhe.borehole.H
            self.calculated_heights[i] = total_drilling

            if old_height < total_drilling:
                break
            else:
                old_height = total_drilling

            i += 1

        keys = list(self.calculated_heights.keys())
        values = list(self.calculated_heights.values())

        minimum_total_drilling = min(values)
        idx = values.index(minimum_total_drilling)
        selection_key_outer = keys[idx]
        self.calculated_temperatures = self.calculated_temperatures_nested[selection_key_outer]

        keys = list(self.calculated_temperatures.keys())
        values = list(self.calculated_temperatures.values())

        negative_excess_values = [v for v in values if v <= 0.0]

        excess_of_interest = max(negative_excess_values)
        idx = values.index(excess_of_interest)
        selection_key = keys[idx]
        selected_coordinates = self.coordinates_domain_nested[selection_key_outer][selection_key]

        self.initialize_ghe(
            selected_coordinates,
            self.max_height,
            field_specifier=self.nested_fieldDescriptors[selection_key_outer][selection_key],
        )
        self.ghe.compute_g_functions(self.min_height, self.max_height)
        self.ghe.size(TimestepType.HYBRID, self.max_height, self.min_height, self.max_eft, self.min_eft)

        return selection_key, selected_coordinates



from math import ceil

import numpy as np
from pygfunction.boreholes import Borehole
from scipy.interpolate import interp1d

from ghedesigner.constants import SEC_IN_HR, TWO_PI, VERSION
from ghedesigner.enums import PipeType, TimestepType
from ghedesigner.ghe.boreholes.factory import get_bhe_object
from ghedesigner.ghe.gfunction import GFunction, calc_g_func_for_multiple_lengths
from ghedesigner.ghe.ground_loads import HybridLoad
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.media import Grout, Soil
from ghedesigner.utilities import combine_sts_lts, solve_root


class GHE:
    def __init__(
        self,
        v_flow_system: float,
        b_spacing: float,
        bhe_type: PipeType,
        fluid,
        borehole: Borehole,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        g_function: GFunction,
        start_month: int,
        end_month: int,
        hourly_extraction_ground_loads: list,
        field_type="N/A",
        field_specifier="N/A",
    ) -> None:
        self.field_type = field_type
        self.fieldSpecifier = field_specifier
        self.v_flow_system = v_flow_system
        self.b_spacing = b_spacing
        self.nbh = len(g_function.bore_locations)
        self.v_flow_borehole = self.v_flow_system / self.nbh
        m_flow_borehole = self.v_flow_borehole / 1000.0 * fluid.rho
        self.m_flow_borehole = m_flow_borehole

        # Borehole Heat Exchanger
        self.bhe_type = bhe_type
        self.bhe = get_bhe_object(bhe_type, m_flow_borehole, fluid, borehole, pipe, grout, soil)

        # Equivalent borehole Heat Exchanger
        self.bhe_eq = self.bhe.to_single()

        # Radial numerical short time step
        self.bhe_eq.calc_sts_g_functions()

        # gFunction object
        self.gFunction = g_function
        # Additional simulation parameters
        self.start_month = start_month
        self.end_month = end_month

        # Hourly ground extraction loads
        # Building cooling is negative, building heating is positive
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.times = np.empty((0,), dtype=np.float64)
        self.loading: list | None = None

        self.hybrid_load = HybridLoad(
            self.hourly_extraction_ground_loads, self.bhe_eq, self.bhe_eq, start_month, end_month
        )

        # List of heat pump exiting fluid temperatures
        self.hp_eft: list[float] = []
        # list of change in borehole wall temperatures
        self.dTb: list[float] = []

    def as_dict(self) -> dict:
        output = {
            "title": f"GHEDesigner GHE Output - Version {VERSION}",
            "number_of_boreholes": len(self.gFunction.bore_locations),
            "borehole_depth": {"value": self.bhe.borehole.H, "units": "m"},
            "borehole_spacing": {"value": self.b_spacing, "units": "m"},
            "borehole_heat_exchanger": self.bhe.as_dict(),
            "equivalent_borehole_heat_exchanger": self.bhe_eq.as_dict(),
            # "simulation_parameters": self.sim_params.as_dict(),
        }
        return output

    def grab_g_function(self, b_over_h):
        # interpolate for the Long time step g-function
        g_function, rb_value, _, _ = self.gFunction.g_function_interpolation(b_over_h)
        # correct the long time step for borehole radius
        g_function_corrected = self.gFunction.borehole_radius_correction(g_function, rb_value, self.bhe.borehole.r_b)
        # Don't Update the HybridLoad (its dependent on the STS) because
        # it doesn't change the results much, and it slows things down a lot
        # combine the short and long time step g-function
        g = combine_sts_lts(
            self.gFunction.log_time,
            g_function_corrected,
            self.bhe_eq.lntts.tolist(),
            self.bhe_eq.g.tolist(),
        )

        g_bhw = combine_sts_lts(
            self.gFunction.log_time,
            g_function_corrected,
            self.bhe_eq.lntts.tolist(),
            self.bhe_eq.g_bhw.tolist(),
        )

        return g, g_bhw

    def cost(self, max_eft: float, min_eft: float, design_max_eft: float, design_min_eft: float):
        delta_t_max = max_eft - design_max_eft
        delta_t_min = design_min_eft - min_eft
        t_excess = max(delta_t_max, delta_t_min)
        return t_excess

    def _simulate_detailed(self, q_dot: np.ndarray, time_values: np.ndarray, g: interp1d):
        # Perform a detailed simulation based on a numpy array of heat rejection
        # rates, Q_dot (Watts) where each load is applied at the time_value
        # (seconds). The g-function can interpolate.
        # Source: Chapter 2 of Advances in Ground Source Heat Pumps

        n = q_dot.size

        # Convert the total load applied to the field to the average over
        # borehole wall rejection rate
        # At time t=0, make the heat rejection rate 0.
        q_dot_b = np.hstack((0.0, q_dot / float(self.nbh)))
        time_values = np.hstack((0.0, time_values))

        q_dot_b_dt = np.hstack(q_dot_b[1:] - q_dot_b[:-1])

        ts = self.bhe_eq.t_s  # (-)
        two_pi_k = TWO_PI * self.bhe.soil.k  # (W/m.K)
        h = self.bhe.borehole.H  # (meters)
        tg = self.bhe.soil.ugt  # (Celsius)
        rb = self.bhe.calc_effective_borehole_resistance()  # (m.K/W)
        m_dot = self.bhe.m_flow_borehole  # (kg/s)
        cp = self.bhe.fluid.cp  # (J/kg.s)

        hp_eft: list[float] = []
        delta_tb: list[float] = []
        for i in range(1, n + 1):
            # Take the last i elements of the reversed time array
            _time = time_values[i] - time_values[0:i]
            # _time = time_values_reversed[n - i:n]
            g_values = g(np.log((_time * SEC_IN_HR) / ts))
            # Tb = Tg + (q_dt * g)  (Equation 2.12)
            delta_tb_i = (q_dot_b_dt[0:i] / h / two_pi_k).dot(g_values)
            # Tf = Tb + q_i * R_b^* (Equation 2.13)
            tb = tg + delta_tb_i
            # Bulk fluid temperature
            tf_bulk = tb + q_dot_b[i] / h * rb
            # T_out = T_f - Q / (2 * m_dot cp)  (Equation 2.14)
            tf_out = tf_bulk - q_dot_b[i] / (2 * m_dot * cp)
            hp_eft.append(tf_out)
            delta_tb.append(delta_tb_i)

        return hp_eft, delta_tb

    def compute_g_functions(self, h_min: float, h_max: float):
        # Compute g-functions for a bracketed solution, based on min and max height
        self.gFunction = calc_g_func_for_multiple_lengths(
            self.b_spacing,
            [h_min] if h_min == h_max else [h_min, (h_min + h_max) / 2.0, h_max],
            self.bhe.borehole.r_b,
            self.bhe.borehole.D,
            self.bhe.m_flow_borehole,
            self.bhe_type,
            self.gFunction.log_time,
            self.gFunction.bore_locations,
            self.bhe.fluid,
            self.bhe.pipe,
            self.bhe.grout,
            self.bhe.soil,
        )

    def simulate(self, method: TimestepType):
        b = self.b_spacing
        b_over_h = b / self.bhe.borehole.H

        # Solve for equivalent single U-tube
        self.bhe_eq = self.bhe.to_single()
        # Update short time step object with equivalent single u-tube
        self.bhe_eq.calc_sts_g_functions()
        # Combine the short and long-term g-functions. The long term g-function
        # is interpolated for specific B/H and rb/H values.
        g, _ = self.grab_g_function(b_over_h)

        if method == TimestepType.HYBRID:
            q_dot = self.hybrid_load.load[2:] * 1000.0  # convert to Watts
            time_values = self.hybrid_load.hour[2:]  # convert to seconds
            self.times = time_values
            self.loading = q_dot

            hp_eft, d_tb = self._simulate_detailed(q_dot, time_values, g)
        elif method == TimestepType.HOURLY:
            n_months = self.end_month - self.start_month + 1
            n_hours = int(n_months / 12.0 * 8760.0)
            q_dot = self.hourly_extraction_ground_loads
            # How many times does q need to be repeated?
            n_years = ceil(n_hours / 8760)
            if len(q_dot) // 8760 < n_years:
                q_dot = q_dot * n_years
            else:
                n_hours = len(q_dot)
            q_dot = -1.0 * np.array(q_dot)  # Convert loads to rejection
            # print("Times:",self.times)
            if len(self.times) == 0:
                self.times = np.arange(1, n_hours + 1, 1)
            t = self.times
            self.loading = q_dot

            hp_eft, d_tb = self._simulate_detailed(q_dot, t, g)
        else:
            raise ValueError("Only hybrid or hourly methods available.")

        self.hp_eft = hp_eft
        self.dTb = d_tb

        return max(hp_eft), min(hp_eft)

    def size(
        self, method: TimestepType, max_height: float, min_height: float, design_max_eft: float, design_min_eft: float
    ) -> None:
        # Size the ground heat exchanger
        def local_objective(h: float):
            self.bhe.borehole.H = h
            this_max_hp_eft, this_min_hp_eft = self.simulate(method=method)
            t_excess = self.cost(this_max_hp_eft, this_min_hp_eft, design_max_eft, design_min_eft)
            return t_excess

        # Make the initial guess variable the average of the heights given
        self.bhe.borehole.H = (max_height + min_height) / 2.0
        # bhe.b.H is updated during sizing
        returned_height = solve_root(
            self.bhe.borehole.H,
            local_objective,
            lower=min_height,
            upper=max_height,
            abs_tol=1.0e-6,
            rel_tol=1.0e-6,
            max_iter=50,
        )

        self.bhe.borehole.H = returned_height

    @staticmethod
    def calculate(_hour_index: int, inlet_temp: float, _flow_rate: float) -> float:
        effectiveness = 0.5
        soil_temp = 20

        return effectiveness * (soil_temp - inlet_temp) + inlet_temp



from ghedesigner.heat_pump import HeatPump


class Building:
    def __init__(self, name: str) -> None:
        self.name = name
        self.heat_pumps: list[HeatPump] = []

    def add_heat_pump(self, heat_pump: HeatPump):
        self.heat_pumps.append(heat_pump)

    def calculate(self, simulation_time: int, inlet_temperature: float, flow_rate: float) -> float:
        outlet_temperature = inlet_temperature
        for heat_pump in self.heat_pumps:
            outlet_temperature = heat_pump.calculate(simulation_time, outlet_temperature, flow_rate)
        return outlet_temperature



import re
from datetime import datetime
from math import floor
from typing import Any

from ghedesigner.ghe.boreholes.base import GHEDesignerBoreholeBase
from ghedesigner.ghe.boreholes.coaxial_borehole import CoaxialPipe
from ghedesigner.ghe.ground_heat_exchangers import GHE
from ghedesigner.output.converters import hours_to_month
from ghedesigner.output.formatters import create_line, create_table, create_title, d_row


class TextSerializer:
    """Generate the plain-text SimulationSummary report."""

    @staticmethod
    def summary_text(
        width: int,
        project_name: str,
        model_name: str,
        notes: str,
        author: str,
        time_sec: float,
        ghe: GHE,
        search_tracker: Any,
    ) -> str:
        # format specifiers
        f_str = "s"
        f_int = ".0f"
        f_1f = ".1f"
        f_2f = ".2f"
        f_3f = ".3f"
        f_4f = ".4f"
        f_sci = ".3e"

        blank = create_line(width)
        empty = create_line(width, character=" ")

        out = blank
        out += d_row(width, "Project Name:", project_name, f_str)
        out += blank
        out += f"Notes:\n\n{notes}\n"
        out += blank
        out += d_row(width, "File/Model Name:", model_name, f_str)
        now = datetime.now().strftime("%m/%d/%Y %H:%M:%S %p")
        out += d_row(width, "Simulated On:", now, f_str)
        out += d_row(width, "Simulated By:", author, f_str)
        out += d_row(width, "Calculation Time, s:", time_sec, f_3f)
        out += empty

        # --- Design Selection Table ---
        out += create_title(width, "Design Selection", filler_symbol="-")
        header = [
            ["Field", "Excess Temperature", "Max Temperature", "Min Temperature"],
            [" ", "(C)", "(C)", "(C)"],
        ]
        col_fmts = [f_str, f_2f, f_2f, f_2f]
        out += create_table(
            "Field Search Log",
            header,
            search_tracker,
            width,
            col_fmts,
            filler_symbol="-",
            centering="^",
        )
        out += empty

        # --- GHE System (gFunction LTS table) ---
        out += create_title(width, "GHE System", filler_symbol="-")
        lts = ghe.gFunction.g_lts
        titles = ["ln(t/ts)"] + [f"H: {h:0.2f} m" for h in lts] + [f"H: {ghe.bhe.borehole.H:0.2f} m"]
        data = []
        interp = ghe.gFunction.g_function_interpolation(ghe.b_spacing / float(ghe.bhe.borehole.H))[0]
        for i, t in enumerate(ghe.gFunction.log_time):
            row = [t] + [lts[h][i] for h in lts] + [interp[i]]
            data.append(row)
        fmts = [f_3f] * len(titles)
        out += create_table("gFunction LTS Values", [titles], data, width, fmts, filler_symbol="-", centering="^")
        out += empty

        # --- System Parameters ---
        out += create_title(width, "System Parameters", filler_symbol="-")
        out += d_row(width, "Active Borehole Length, m:", ghe.bhe.borehole.H, f_int)
        out += d_row(width, "Borehole Diameter, mm:", ghe.bhe.borehole.r_b * 2e3, f_2f)
        out += d_row(width, "Borehole Spacing, m:", ghe.b_spacing, f_3f)
        out += d_row(width, "Borehole Depth, m:", ghe.bhe.borehole.D, f_2f)
        out += d_row(
            width,
            "Total Drilling, m:",
            ghe.bhe.borehole.H * len(ghe.gFunction.bore_locations),
            f_int,
        )
        out += "Field Geometry: \n"
        out += d_row(width, "Field Type:", ghe.field_type, f_str, n_tabs=1)
        out += d_row(width, "Field Specifier:", ghe.fieldSpecifier, f_str, n_tabs=1)
        out += d_row(width, "NBH:", len(ghe.gFunction.bore_locations), f_int, n_tabs=1)

        # --- Borehole Information ---
        out += "Borehole Information: \n"
        # pipe geometry
        if isinstance(ghe.bhe.pipe.r_out, float):
            out += d_row(width, "Pipe Outer Diameter, mm:", ghe.bhe.pipe.r_out * 2e3, f_2f, n_tabs=1)
            out += d_row(width, "Pipe Inner Diameter, mm:", ghe.bhe.pipe.r_in * 2e3, f_2f, n_tabs=1)
        else:
            out += d_row(width, "Outer Pipe Outer Diameter, mm:", ghe.bhe.pipe.r_out[1] * 2e3, f_2f, n_tabs=1)
            out += d_row(width, "Outer Pipe Inner Diameter, mm:", ghe.bhe.pipe.r_out[0] * 2e3, f_2f, n_tabs=1)
            out += d_row(width, "Inner Pipe Outer Diameter, mm:", ghe.bhe.pipe.r_in[1] * 2e3, f_2f, n_tabs=1)
            out += d_row(width, "Inner Pipe Inner Diameter, mm:", ghe.bhe.pipe.r_in[0] * 2e3, f_2f, n_tabs=1)

        out += d_row(width, "Pipe Roughness, m:", ghe.bhe.pipe.roughness, f_sci, n_tabs=1)
        if isinstance(ghe.bhe.pipe.k, float):
            out += d_row(width, "Pipe Thermal Conductivity, W/m-K:", ghe.bhe.pipe.k, f_3f, n_tabs=1)
        else:
            out += d_row(width, "Inner Pipe Thermal Conductivity, W/m-K:", ghe.bhe.pipe.k[0], f_3f, n_tabs=1)
            out += d_row(width, "Outer Pipe Thermal Conductivity, W/m-K:", ghe.bhe.pipe.k[1], f_3f, n_tabs=1)
        out += d_row(width, "Pipe Volumetric Heat Capacity, kJ/m3-K:", ghe.bhe.pipe.rho_cp / 1000, f_2f, n_tabs=1)
        out += d_row(width, "Shank Spacing, mm:", ghe.bhe.pipe.s * 1e3, f_2f, n_tabs=1)
        out += d_row(width, "Grout Thermal Conductivity, W/(m-K):", ghe.bhe.grout.k, f_3f, n_tabs=1)
        out += d_row(width, "Grout Volumetric Heat Capacity, kJ/m3-K:", ghe.bhe.grout.rho_cp / 1000, f_2f, n_tabs=1)

        # Reynolds & effective resistance
        if isinstance(ghe.bhe.pipe.r_out, float):
            reynolds = GHEDesignerBoreholeBase.compute_reynolds(
                ghe.bhe.m_flow_borehole, ghe.bhe.pipe.r_in, ghe.bhe.fluid
            )
        else:
            reynolds = CoaxialPipe.compute_reynolds_concentric(
                ghe.bhe.m_flow_borehole,
                ghe.bhe.r_in_out,
                ghe.bhe.r_out_in,
                ghe.bhe.fluid,
            )
        out += d_row(width, "Reynold's Number:", reynolds, f_int, n_tabs=1)
        out += d_row(
            width,
            "Effective Borehole Resistance, K/(W/m):",
            ghe.bhe.calc_effective_borehole_resistance(),
            f_4f,
            n_tabs=1,
        )

        # Soil & Fluid
        out += "Soil Properties: \n"
        out += d_row(width, "Thermal Conductivity, W/m-K:", ghe.bhe.soil.k, f_3f, n_tabs=1)
        out += d_row(width, "Volumetric Heat Capacity, kJ/m3-K:", ghe.bhe.soil.rho_cp / 1000, f_2f, n_tabs=1)
        out += d_row(width, "Undisturbed Ground Temperature, C:", ghe.bhe.soil.ugt, f_2f, n_tabs=1)

        out += "Fluid Properties\n"
        out += d_row(width, "Volumetric Heat Capacity, kJ/m3-K:", ghe.bhe.fluid.rho_cp / 1000, f_2f, n_tabs=1)
        out += d_row(width, "Thermal Conductivity, W/m-K:", ghe.bhe.fluid.k, f_2f, n_tabs=1)
        out += d_row(width, "Viscosity, Pa-s:", ghe.bhe.fluid.mu, f_sci, n_tabs=1)
        out += d_row(width, "Fluid Mix:", ghe.bhe.fluid.name, f_str, n_tabs=1)
        out += d_row(width, "Density, kg/m^3:", ghe.bhe.fluid.rho, f_2f, n_tabs=1)
        out += d_row(width, "Mass Flow Rate Per Borehole, kg/s:", ghe.bhe.m_flow_borehole, f_3f, n_tabs=1)
        if hasattr(ghe.bhe, "h_f"):
            out += d_row(width, "Fluid Convection Coefficient, W/m-K:", ghe.bhe.h_f, f_int, n_tabs=1)
        out += empty

        # --- Hybrid load monthly tables & simulation results (if present) ---
        if getattr(ghe, "hybrid_load", None):
            monthly_load_values: list[list[Any]] = []
            n_months = len(ghe.hybrid_load.monthly_cl) - 1
            n_years = int(n_months / 12)
            months = n_years * [
                "January",
                "February",
                "March",
                "April",
                "May",
                "June",
                "July",
                "August",
                "September",
                "October",
                "November",
                "December",
            ]

            for i in range(1, n_months + 1):
                hl = ghe.hybrid_load
                monthly_load_values.append(
                    [
                        months[i - 1],
                        hl.monthly_hl[i],
                        hl.monthly_cl[i],
                        hl.monthly_peak_hl[i],
                        hl.monthly_peak_hl_duration[i],
                        hl.monthly_peak_cl[i],
                        hl.monthly_peak_cl_duration[i],
                    ]
                )
            month_header = [
                [
                    "Month",
                    "Total Heating",
                    "Total Cooling",
                    "Peak Heating",
                    "PH Duration",
                    "Peak Cooling",
                    "PC Duration",
                ],
                ["", "kWh", "kWh", "kW", "hr", "kW", "hr"],
            ]
            month_table_formats = [f_str, f_1f, f_1f, f_1f, f_1f, f_1f, f_1f]
            out += create_table(
                "GLHE Monthly Loads",
                month_header,
                monthly_load_values,
                width,
                month_table_formats,
                filler_symbol="-",
                centering="^",
            )
            out += empty

            # Simulation Parameters header (original code left details commented)
            out += create_title(width, "Simulation Parameters")
            out += empty

            # Simulation Results (peaks + monthly temp summary)
            out += create_title(width, "Simulation Results")
            out += empty

            # Peak Temperature section
            max_eft = max(ghe.hp_eft)
            min_eft = min(ghe.hp_eft)
            max_eft_time_hrs = ghe.times[ghe.hp_eft.index(max_eft)]
            min_eft_time_hrs = ghe.times[ghe.hp_eft.index(min_eft)]
            out += create_title(width, "Peak Temperature", filler_symbol="-")
            out += d_row(width, "Max HP EFT, C:", max_eft, f_3f)
            out += d_row(width, "Max HP EFT Time, Months:", hours_to_month(max_eft_time_hrs), f_3f)
            out += d_row(width, "Min HP EFT, C:", min_eft, f_3f)
            out += d_row(width, "Min HP EFT Time, Months:", hours_to_month(min_eft_time_hrs), f_3f)

            # Monthly temperature summary table
            out_array = []
            last_month = -1
            month_tb_vals: list[float] = []
            month_eft_vals: list[float] = []
            for tv, d_tb, eft in zip(ghe.times, ghe.dTb, ghe.hp_eft):
                current_month = floor(hours_to_month(tv))
                if current_month == last_month:
                    month_tb_vals.append(d_tb)
                    month_eft_vals.append(eft)
                else:
                    if month_tb_vals:
                        previous_temp = ghe.bhe.soil.ugt
                        out_array.append(
                            [
                                current_month,
                                previous_temp + month_tb_vals[-1],
                                max(month_eft_vals),
                                min(month_eft_vals),
                            ]
                        )
                    last_month = current_month
                    month_tb_vals = [d_tb]
                    month_eft_vals = [eft]

            header_array = [
                ["Time", "BH Wall Temp", "Max HP EFT", "Min HP EFT"],
                ["(months)", "(C)", "(C)", "(C)"],
            ]
            eft_table_formats = [f_int, f_2f, f_2f, f_2f]
            out += create_table(
                "Monthly Temperature Summary",
                header_array,
                out_array,
                width,
                eft_table_formats,
                filler_symbol="-",
                centering="^",
            )

        # strip trailing spaces
        return re.sub(r"[ \t]+\n", "\n", out)



from queue import Queue
from tkinter import (
    BOTH,
    EW,
    LEFT,
    NSEW,
    SUNKEN,
    TOP,
    Button,
    Frame,
    Label,
    Menu,
    S,
    StringVar,
    Tk,
    X,
)
from tkinter.ttk import LabelFrame  # ttk widgets
from typing import Any


class GHEDesignerWindow(Tk):
    """This form is the primary GUI entry point for the program; all control runs through here"""

    def __init__(self) -> None:
        """
        The main window of the parameter estimation tool GUI workflow.
        This window is an instance of a tk.Tk object
        """
        super().__init__(className="GHEDesignerWindow")

        # set some basic program information like title and an icon
        self.title("GHEDesignerWindow")

        # setup event listeners
        self._gui_queue: Queue[Any] = Queue()
        self._check_queue()

        # define the Tk.Variable instances that will be used to communicate with the GUI widgets
        self._define_tk_variables()

        # build out the form and specify a minimum size, which may not be uniform across platforms
        self._build_gui()
        self.update()
        self.minsize(self.winfo_width(), self.winfo_height())

        # set up some important member variables
        self._thread_running = False

        # window setup operations
        self._update_status_bar("Program Initialized")
        self._refresh_gui_state()
        self.bind("<Key>", self._handle_button_pressed)

    def _handle_button_pressed(self, event):
        # relevant_modifiers
        # mod_shift = 0x1
        mod_control = 0x4
        # mod_alt = 0x20000
        if event.keysym == "e" and mod_control & event.state:
            pass

    def _check_queue(self):
        """Checks the GUI queue for actions and sets a timer to check again each time"""
        while True:
            try:
                task = self._gui_queue.get(block=False)
                self.after_idle(task, [])
            except Exception:  # noqa: BLE001
                break
        self.after(100, self._check_queue)

    def _define_tk_variables(self):
        """Creates and initializes all the Tk.Variable instances used in the GUI for two-way communication"""
        self._tk_var_status = StringVar(value="Program Initialized")

    def _build_gui(self):
        """Builds out the entire window GUI, calling workers as necessary"""
        # now build the top menubar, it's not part of the geometry, but a config parameter on the root Tk object
        menubar = Menu(self)
        menu_exit = Menu(menubar, tearoff=0)
        menubar.add_cascade(label="Exit", menu=menu_exit)
        self.config(menu=menubar)

        # main contents
        label_frame_control = LabelFrame(self, text="Stuff")
        Label(label_frame_control, text="Here is where we could put some stuff").pack(side=TOP, padx=3, pady=3)
        self._button_engage = Button(label_frame_control, text="Button", command=self._button)
        self._button_engage.pack(side=TOP, padx=3, pady=3, fill=X)
        label_frame_control.grid(row=0, column=0, sticky=NSEW)

        # build the status bar
        status_frame = Frame(self)
        Label(status_frame, relief=SUNKEN, anchor=S, textvariable=self._tk_var_status).pack(
            side=LEFT, fill=BOTH, expand=True
        )
        status_frame.grid(row=1, column=0, columnspan=3, sticky=EW)

        # set up the weight of each row (even) and column (distributed) for a nice looking GUI
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)

    def _button(self):
        pass

    def _refresh_gui_state(self):
        """Checks current instance flags and sets button states appropriately"""
        if self._thread_running:
            pass
        else:
            pass

    def _update_status_bar(self, extra_message: str) -> None:
        """
        Updates the status bar at the bottom of the window, providing data based on flags and displaying the message

        :param extra_message: String message to show on the right side of the status bar
        :return: Nothing
        """
        self._tk_var_status.set(extra_message)

    def run(self) -> None:
        """Executes the Tk main loop to handle all GUI events and update"""
        self.mainloop()



from functools import lru_cache
from math import atan, cos, inf, pi, sin, sqrt

import numpy as np

from ghedesigner.constants import DEG_TO_RAD, PI_OVER_2, RAD_TO_DEG
from ghedesigner.ghe.shape import Shapes, point_polygon_check, sort_intersections

NEIGHBORHOOD_OFFSETS = ((-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 0), (0, 1), (1, -1), (1, 0), (1, 1))


class DeferredDuplicateCheckList:
    def __init__(self, spacing):
        self.points = {}
        self.buckets = {}
        self.points_to_check = []
        self.spacing = spacing
        self.partitioned = False
        self.proximity_checks_found = False
        self.largest_index = 0
        self.bucket_keys = None

    def _get_bucket_key(self, px, py):
        return int(px // self.spacing), int(py // self.spacing)

    def _get_bucket_keys(self, points_array):
        return (points_array / self.spacing).astype(int)

    def size(self):
        return len(self.points)

    def delete(self, bucket_keys_index):
        key = self.bucket_keys[bucket_keys_index]
        key = (key[0], key[1])
        point_keys = list(self.points.keys())
        point_index = point_keys[bucket_keys_index]
        index_to_delete = self.buckets[key].index(point_index)
        del self.buckets[key][index_to_delete]
        del self.points[point_index]
        del self.bucket_keys[bucket_keys_index]

    def index(self, ind):
        keys = list(self.points.keys())
        return self.points[keys[ind]]

    def partition(self):
        array_repr = self.toarray()
        bucket_keys = self._get_bucket_keys(array_repr)
        self.bucket_keys = bucket_keys.tolist()
        point_keys = list(self.points.keys())
        for i, key in enumerate(bucket_keys):
            usable_key = (key[0], key[1])
            if usable_key in self.buckets:
                self.buckets[usable_key].append(point_keys[i])
            else:
                self.buckets[usable_key] = [point_keys[i]]

        self.partitioned = True

    def find_proximity_checks(self):
        if not self.partitioned:
            self.partition()
        for i, key in enumerate(self.points):
            x_ind, y_ind = self.bucket_keys[i]
            for key_modifier in NEIGHBORHOOD_OFFSETS:
                dx, dy = key_modifier
                neighbor_key = (x_ind + dx, y_ind + dy)
                for j in self.buckets.get(neighbor_key, []):
                    if j <= key:
                        continue
                    self.points_to_check.append((key, j))
        self.proximity_checks_found = True

    def append(self, element):
        # Since the spatial partitioning is done in bulk in "partition()", we have to reset the partitioning
        # when appending a new element.
        if self.partitioned:
            self.partitioned = False
            self.buckets = {}
            self.bucket_keys = None
        # For a similar reason, the proximity checks need to be rerun when a new element is appended.
        if self.proximity_checks_found:
            self.proximity_checks_found = False
            self.points_to_check = []
        # Internally, points are contained in the self.points dictionary, but we want this classes usage
        # to be similar to a list/array. This can cause an issue when doing deletions (as we would
        # like to avoid readjusting the keys for all dict entries). In order to avoid generating an already
        # existing key, "largest_index" is used to always increment the key for the added element (regardless
        # of deletions). There is probably a better way to implement this internally, as I suspect the
        # point indexing with this implementation is slower than necessary.
        self.points[self.largest_index] = element
        self.largest_index += 1

    def extend(self, elements):
        for element in elements:
            self.append(element)

    def tolist(self):
        return list(self.points.values())

    def toarray(self):
        return np.array(self.tolist(), dtype=float)

    def find_duplicates(self, tolerance):
        if tolerance > self.spacing:
            raise ValueError("Requested tolerance exceeds that which is allowed by the given spatial partitioning.")

        if not self.proximity_checks_found:
            self.find_proximity_checks()

        duplicates = []
        keys_list = list(self.points.keys())
        squared_tolerance = tolerance * tolerance
        for points_to_check in self.points_to_check:
            i, j = points_to_check
            p1 = self.points[i]
            p2 = self.points[j]
            dx = p1[0] - p2[0]
            dy = p1[1] - p2[1]
            test_dist = dx * dx + dy * dy
            if test_dist < squared_tolerance:
                duplicates.append((keys_list.index(i), keys_list.index(j)))
        return duplicates

    def get_line_partitions(self, p1x, p1y, p2x, p2y):
        # This is based on Amanatides & Woo's algorithm as described here:
        # https://github.com/cgyurgyik/fast-voxel-traversal-algorithm/blob/master/overview/FastVoxelTraversalOverview.md
        # It should be noted that this implementation assumes that the line segment begins on the spatial grid
        # which simplifies the initialization of some of the values.

        buckets_visited = []

        p1x_bucket, p1y_bucket = self._get_bucket_key(p1x, p1y)
        p2x_bucket, p2y_bucket = self._get_bucket_key(p2x, p2y)

        dx = p2x - p1x
        dy = p2y - p1y

        if dx == 0:
            step_x = 0
            t_max_x = float("inf")
        elif dx > 0:
            step_x = 1
            t_max_x = ((p1x_bucket + 1) * self.spacing - p1x) / dx
        else:
            step_x = -1
            t_max_x = (p1x_bucket * self.spacing - p1x) / dx

        if dy == 0:
            step_y = 0
            t_max_y = float("inf")
        elif dy > 0:
            step_y = 1
            t_max_y = ((p1y_bucket + 1) * self.spacing - p1y) / dy
        else:
            step_y = -1
            t_max_y = (p1y_bucket * self.spacing - p1y) / dy

        t_delta_x = self.spacing / abs(dx) if dx != 0 else float("inf")
        t_delta_y = self.spacing / abs(dy) if dy != 0 else float("inf")

        current_x_bucket = p1x_bucket
        current_y_bucket = p1y_bucket
        while True:
            buckets_visited.append((current_x_bucket, current_y_bucket))
            if current_x_bucket == p2x_bucket and current_y_bucket == p2y_bucket:
                break
            if t_max_x < t_max_y:
                current_x_bucket += step_x
                t_max_x += t_delta_x
            else:
                current_y_bucket += step_y
                t_max_y += t_delta_y
        return buckets_visited

    def points_close_to_line(self, p1, p2, tolerance):
        if tolerance > self.spacing:
            raise ValueError("Requested tolerance exceeds that which is allowed by the given spatial partitioning.")

        p1x, p1y = p1
        p2x, p2y = p2

        # Trace path of line segment through partitioned space
        buckets_to_check = self.get_line_partitions(p1x, p1y, p2x, p2y)

        # Add in neighboring regions
        neighbors_to_check = []
        for bucket in buckets_to_check:
            for key_modifier in NEIGHBORHOOD_OFFSETS:
                other_bucket = (bucket[0] + key_modifier[0], bucket[1] + key_modifier[1])
                neighbors_to_check.append(other_bucket)

        # Return points that might be too close.
        keys_list = list(self.points.keys())
        buckets_to_check.extend(neighbors_to_check)
        buckets_to_check = set(buckets_to_check)
        close_points = []
        for bucket in buckets_to_check:
            for elem in self.buckets.get(bucket, []):
                close_points.append(keys_list.index(elem))

        return close_points


def gen_shape(prop_bound: list[list[float]], ng_zones=None):
    """Returns an array of shapes objects representing the coordinates given"""
    r_a: list[Shapes | list[Shapes] | None] = [Shapes(prop_bound)]
    if ng_zones is not None:
        r_n = []
        for ng_zone in ng_zones:
            r_n.append(Shapes(ng_zone))
        r_a.append(r_n)
    else:
        r_a.append(None)
    return r_a


def field_optimization_wp_space_fr(
    p_space,
    space_start,
    rotate_step,
    prop_bound,
    ng_zones=None,
    rotate_start=None,
    rotate_stop=None,
    partition_ratio=1.0,
    duplicate_spacing_ratio=0.1,
):
    """Optimizes a Field by iterating over input values w/o perimeter spacing

    Parameters: p_space(float): Ratio of perimeter spacing to other target spacing space_start(float): the initial
    target spacing that the optimization program will start with rotate_step(float): the amount of rotation that will
    be changed per step (in degrees) prop_bound([[float,float]]): 2d array of floats that represent the property
    boundary (counterclockwise) ng_zones([[[float,float]]]): 3d array representing the different zones on the
    property where no boreholes can be placed rotate_start(float): the rotation that the field will start at (-pi/2 <
    rotateStart < pi/2) rotate_stop(float): the rotation that the field will stop at (exclusive) (-pi/2 < rotateStop
    < pi/2)

    Outputs: CSVs containing the coordinates for the max field for each target spacing, their respective graphs,
    and their respective data

    """
    if rotate_start is None:
        rotate_start = (-90.0 + rotate_step) * DEG_TO_RAD
    if rotate_stop is None:
        rotate_stop = PI_OVER_2
    if rotate_start > PI_OVER_2 or rotate_start < -PI_OVER_2 or rotate_stop > PI_OVER_2 or rotate_stop < -PI_OVER_2:
        raise ValueError("Invalid Rotation")

    space = space_start
    rt = rotate_start

    y_s = space
    x_s = y_s

    max_l = 0
    max_hole = None
    max_rt = None

    while rt < rotate_stop:
        hole = two_space_gen_bhc(
            prop_bound,
            y_s,
            x_s,
            rotate=rt,
            no_go=ng_zones,
            p_space=p_space * x_s,
            intersection_tolerance=1e-5,
            duplicate_spacing_ratio=duplicate_spacing_ratio,
            partition_ratio=partition_ratio,
        )

        # Assuming that the rotation with the maximum number of boreholes is most efficiently using space
        if hole.size() > max_l:
            max_l = hole.size()
            max_rt = rt * RAD_TO_DEG
            max_hole = hole

        rt += rotate_step * DEG_TO_RAD

    # Ensures that there are no repeated boreholes
    if duplicate_spacing_ratio != 0:
        max_hole = remove_duplicates(max_hole, duplicate_spacing_ratio * max(x_s, y_s, p_space * x_s))

    field = max_hole
    field_name = f"P{p_space:0.1f}_S{space:0.1f}_rt{max_rt:0.1f}"
    return [field.toarray(), field_name]


def field_optimization_fr(
    space_start,
    rotate_step,
    prop_bound,
    ng_zones=None,
    rotate_start=None,
    rotate_stop=None,
    intersection_tolerance=1e-5,
    partition_ratio=1.0,
    duplicate_spacing_ratio=0.1,
):
    """Optimizes a Field by iterating over input values w/o perimeter spacing

    Parameters: space_start(float): the initial target spacing that the optimization program will start with
    rotate_step(float): the amount of rotation that will be changed per step (in degrees) prop_bound([[float,
    float]]): 2d array of floats that represent the property boundary (counterclockwise) ng_zones([[[float,
    float]]]): 3d array representing the different zones on the property where no boreholes can be placed
    rotate_start(float): the rotation that the field will start at (-pi/2 < rotateStart < pi/2) rotate_stop(float):
    the rotation that the field will stop at (exclusive) (-pi/2 < rotateStop < pi/2) intersection_tolerance:

    Outputs: CSVs containing the coordinates for the max field for each target spacing, their respective graphs,
    and their respective data

    """
    if rotate_start is None:
        rotate_start = -90.0 * DEG_TO_RAD
    if rotate_stop is None:
        rotate_stop = PI_OVER_2
    if rotate_start > PI_OVER_2 or rotate_start < -PI_OVER_2 or rotate_stop > PI_OVER_2 or rotate_stop < -PI_OVER_2:
        raise ValueError("Invalid Rotation")

    # Target Spacing iterates

    space = space_start
    rt = rotate_start

    y_s = space
    x_s = y_s

    max_l = 0
    max_hole = None
    max_rt = None

    while rt < rotate_stop:
        hole = gen_borehole_config(
            prop_bound,
            y_s,
            x_s,
            rotate=rt,
            no_go=ng_zones,
            intersection_tolerance=intersection_tolerance,
            duplicate_spacing_ratio=duplicate_spacing_ratio,
            partition_ratio=partition_ratio,
        )

        # Assuming that the rotation with the maximum number of boreholes is most efficiently using space
        """
        print(rt)
        import matplotlib.pyplot as plt
        x_vals = [hole.points[point][0] for point in hole.points]
        y_vals = [hole.points[point][1] for point in hole.points]
        plt.scatter(x_vals, y_vals)
        plt.show()
        """
        if hole.size() > max_l:
            max_l = hole.size()
            max_rt = rt * RAD_TO_DEG
            max_hole = hole

        rt += rotate_step * DEG_TO_RAD

    # Ensures that there are no repeated boreholes
    if duplicate_spacing_ratio != 0:
        max_hole = remove_duplicates(max_hole, duplicate_spacing_ratio * max(x_s, y_s))

    field = max_hole
    field_name = f"S_{space:0.1f}_rt{max_rt:0.1f}"
    return [field.toarray(), field_name]


def sum_sq_dist(p1, p2):
    """Returns the **sum of squared** cartesian distance between two points"""
    return (p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2


def pts_dist(p1, p2):
    """Returns the cartesian distance between two points"""
    return sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)


def remove_duplicates(borefield: DeferredDuplicateCheckList, space, disp=False):
    """
    Removes all the duplicates found from the duplicate pairs returned in
    :func:`check_duplicates`.
    For each pair of duplicates, the first borehole (with the lower index) is
    kept and the other (with the higher index) is removed.
    Parameters
    ----------
    borefield : DeferredDuplicateCheckList
        A list of :class:`Borehole` objects
    space:
    disp : bool, optional
        Set to true to print progression messages.
        Default is False.
    Returns
    -------
    new_borefield : DeferredDuplicateCheckList
        A boreField without duplicates
    """
    # get a list of tuple
    original_size = borefield.size()
    duplicate_pairs = borefield.find_duplicates(space)

    # values not to be included
    duplicate_bores = []
    for i in range(len(duplicate_pairs)):
        duplicate_bores.append(duplicate_pairs[i][1])

    duplicate_bores = sorted(set(duplicate_bores), reverse=True)

    for index in duplicate_bores:
        borefield.delete(index)
    if disp:
        # pad with '-' align in center
        print(
            f"{'*gt.boreholes.remove_duplicates()*':-^50}".replace("*", " ")
        )  # keep a space between the function name
        n_duplicates = original_size - borefield.size()
        print(f"The number of duplicates removed: {n_duplicates}")

    return borefield


def two_space_gen_bhc(
    field,
    y_space,
    x_space,
    no_go=None,
    rotate=0,
    p_space=None,
    i_space=None,
    intersection_tolerance=1e-5,
    duplicate_spacing_ratio=0.1,
    partition_ratio=1.0,
) -> np.array:
    """Generates a borefield that has perimeter spacing

    Parameters:
        field: The outer boundary of the property represented as an array of points
        y_space: Target Spacing in y-dir
        x_space: Target Spacing in x-dir
        no_go: a 3d array representing all the areas where boreholes cannot be placed
        rotate: the amount of rotation (rad)
        p_space: Perimeter spacing
        i_space: Spacing required between perimeter and the rest
        i_space: Min spacing required from all edges
        intersection_tolerance:
        duplicate_spacing_ratio: Used for spatial partitioning (likely does not need to be modified).

    """
    if p_space is None:
        p_space = 0.9 * x_space
    if i_space is None:
        i_space = x_space

    # calls the standard row-wise coord generator w/ the adjusted vertices
    holes = gen_borehole_config(
        field,
        y_space,
        x_space,
        no_go=no_go,
        rotate=rotate,
        intersection_tolerance=intersection_tolerance,
        duplicate_spacing_ratio=duplicate_spacing_ratio,
        partition_ratio=partition_ratio,
    )

    remove_points_too_close(field, holes, i_space, no_go_zones=no_go)

    # places the boreholes along the perimeter of the property boundary and no_go zone(s)
    perimeter_distribute(field, p_space, holes)
    if no_go is not None:
        for ng in no_go:
            perimeter_distribute(ng, p_space, holes)
    holes.partition()
    for i in range(holes.size() - 1, -1, -1):
        point = holes.index(i)
        if point_polygon_check(field.c, point, on_edge_tolerance=1e-3) == -1:
            holes.delete(i)
            continue
        if no_go is not None:
            for ng in no_go:
                if point_polygon_check(ng.c, point, on_edge_tolerance=1e-3) == 1:
                    holes.delete(i)
                    break

    return holes


def remove_points_too_close(field, holes, i_space, no_go_zones=None):
    """
    Will remove all points too close to the field and no-go zones

    Parameters:
        field: The outer boundary of the property represented as an array of points
        no_go_zones: a 3d array representing all the areas where boreholes cannot be placed
        holes: 2d array containing all the current boreholes
        i_space: Min spacing required from all edges
    """
    field = field.c
    len_field = len(field)
    for i in range(len_field):
        p1 = field[i]
        p2 = field[0] if i == len_field - 1 else field[i + 1]
        remove_points_close_to_line(p1, p2, holes, i_space)
    if no_go_zones is not None:
        len_no_go_zones = len(no_go_zones)
        for i in range(len_no_go_zones):
            ng = no_go_zones[i].c
            len_ng = len(ng)
            for j in range(len_ng):
                p1 = ng[j]
                p2 = ng[0] if j == len_ng - 1 else ng[j + 1]
                remove_points_close_to_line(p1, p2, holes, i_space)


def remove_points_close_to_line(p1, p2, holes, i_space):
    """Removes points that are close to the given line

    Parameters:
        p1([float,float]): first point in line
        p2([float,float]): second point in line
        holes: 2d array containing a bunch of points
        i_space(float): distance cutoff for how close points can be

    """
    points_to_check = holes.points_close_to_line(p1, p2, i_space)
    points_to_check = sorted(set(points_to_check), reverse=True)
    for point in points_to_check:
        hole = holes.index(point)
        dp = dist_from_line(p1, p2, hole)
        if dp < i_space:
            holes.delete(point)


def dist_from_line(p1, p2, other_point):
    """Calculates the distance from a point to a line (closest distance):
    https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line

    Parameter:
        p1: first point on the line
        p2: second point on the line
        other_point: point which is being measured to

    """
    dxl = p2[0] - p1[0]
    dyl = p2[1] - p1[1]
    dx = p1[0] - other_point[0]
    dy = p1[1] - other_point[1]
    num = abs(dxl * dy - dx * dyl)
    den = sqrt(dxl * dxl + dyl * dyl)
    dp = num / den
    dist_l = pts_dist(p1, p2)
    d01 = pts_dist(p1, other_point)
    d02 = pts_dist(p2, other_point)
    if d01 * d01 - dp * dp < 0:
        return d01
    if sqrt(d01 * d01 - dp * dp) / dist_l > 1:
        return min(d01, d02)
    if min(p1[0], p2[0]) < other_point[0] < max(p1[0], p2[0]) or min(p1[1], p2[1]) < other_point[1] < max(p1[1], p2[1]):
        return min(d01, d02, dp)
    else:
        return min(d01, d02)


def perimeter_distribute(field, space, r):
    """
    Distributes boreholes along the perimeter of a given shape

    Parameters:
        field: array of points representing closed polygon
        space (float): spacing that the boreholes should have from one another
        r (dict{}): existing dictionary of boreholes which will be appended to
    """
    total_arc_length = field.total_arc_length
    act_number_of_boreholes = int(total_arc_length // space)
    perimeter_spacing_ratios = np.linspace(0.0, 1.0, act_number_of_boreholes, endpoint=False)
    for ratio in perimeter_spacing_ratios:
        r.append(field.get_point_along_curve(ratio))


def gen_borehole_config(
    field,
    y_space,
    x_space,
    no_go=None,
    rotate=0,
    intersection_tolerance=1e-6,
    duplicate_spacing_ratio=0.1,
    partition_ratio=1.0,
):
    """
    Function generates a series of x,y points representing a field of boreholes
    in a trapezoidal shape. Returns empty if boreHole field does not meet given requirements

    Parameters
    -------------
        :param field:
        :param y_space: float
            the minimum spacing between points in the y-dir
        :param x_space: float
            the minimum spacing between points in the x-dir
        :param no_go:
        :param rotate:
        :param intersection_tolerance:
        :param duplicate_spacing_ratio
        :param partition_ratio: float
            Defines the size of the space partitioning

        :return: [[float]] -> 2 col + n rows
    """

    if no_go is None:
        no_go = []
    # Decides which vertex to start generating boreholes at by finding the "lowest" vertex relative to a rotated x-axis
    lowest_vert_val = inf
    highest_vert_val = -inf
    lowest_vert = None
    highest_vert = None
    for vert in field.c:
        phi = atan(vert[1] / vert[0]) if vert[0] != 0 else PI_OVER_2
        dist_vert = sqrt(vert[1] ** 2 + vert[0] ** 2)
        ref_angle = phi
        if phi > PI_OVER_2:
            if phi > pi:
                ref_angle = 2 * rotate + 3 * PI_OVER_2 - phi if phi > 3 * PI_OVER_2 else 2 * rotate + pi - phi
            else:
                ref_angle = pi - phi + 2 * rotate
        yp = dist_vert * sin(ref_angle - rotate)
        if yp < lowest_vert_val:
            lowest_vert_val = yp
            lowest_vert = vert
        if yp > highest_vert_val:
            highest_vert_val = yp
            highest_vert = vert

    # Determines the number of rows as well as the distance between the rows
    num_rows = int((highest_vert_val - lowest_vert_val) // y_space)
    d = highest_vert_val - lowest_vert_val
    s = d / num_rows
    row_space = [-1 * s * cos(PI_OVER_2 - rotate), s * sin(PI_OVER_2 - rotate)]

    # Establishes the list object
    boreholes = DeferredDuplicateCheckList(partition_ratio * max(x_space, y_space))
    row_point = [lowest_vert[0], lowest_vert[1]]

    # This is just a value that is combined with the slope of the row's to establish two points defining a row (could
    # be any value)
    point_shift = 1000.0

    for _ in range(num_rows + 1):
        # Row Defined by two points
        if row_space[1] == 0:
            row = [
                row_point[0],
                row_point[1],
                row_point[0],
                row_point[1] + point_shift,
            ]
        else:
            row = [
                row_point[0],
                row_point[1],
                row_point[0] + point_shift,
                row_point[1] + (-row_space[0] / row_space[1]) * point_shift,
            ]

        # Gets Intersection between current row and property boundary
        f_inters = field.line_intersect(row, rotate, intersection_tolerance)

        # Stores the number of intersections with the row
        len_f_inters = len(f_inters)

        # Checks for edge case where a single intersection is reported as two and treats it as one
        if (
            len_f_inters > 1
            and abs(f_inters[0][0] - f_inters[1][0]) <= intersection_tolerance
            and abs(f_inters[0][1] - f_inters[1][1]) <= intersection_tolerance
        ):
            fi = 0
            fij = 0
            while fi < len_f_inters:
                while fij < len_f_inters:
                    if fij == fi:
                        fij += 1
                        continue
                    if (
                        abs(f_inters[fi][0] - f_inters[fij][0]) <= intersection_tolerance
                        and abs(f_inters[fi][1] - f_inters[fij][1]) <= intersection_tolerance
                    ):
                        f_inters.pop(fij)
                        if fi >= fij:
                            fi -= 1
                        fij -= 1
                        len_f_inters -= 1
                    fij += 1
                fi += 1

        # Checks for edge case where there are no intersections detected due to a rounding error (can sometimes
        # happen with the last row)
        """

        if len_f_inters == 0 and ri == num_rows:
            ins = False

            # Checks if the predicted point (ghost point that was expected but not found) is inside one of the no_go
            zones for shape in no_go: if shape.point_intersect(highest_vert): ins = True if not ins: #Double checks
            that this borehole has not already been included if len(boreholes)==0 or not (boreholes[len(boreholes) -
            1][0] == highest_vert[0] and boreholes[len(boreholes) - 1][1] ==highest_vert[1]): boreholes[len(
            boreholes)] = highest_vert """
        # Handles cases with odd number of intersections
        if len_f_inters % 2 == 0:
            # Specific case with two intersections
            if len_f_inters == 2:  # noqa: PLR2004, SIM102
                # Checks for the edge case where two intersections are very close together and replaces them with one
                # point
                if (
                    sqrt(
                        (f_inters[0][0] - f_inters[1][0]) * (f_inters[0][0] - f_inters[1][0])
                        + (f_inters[0][1] - f_inters[1][1]) * (f_inters[0][1] - f_inters[1][1])
                    )
                    < x_space
                ):
                    ins = False
                    for ng_shape in no_go:
                        if ng_shape.point_intersect(highest_vert):
                            ins = True
                    if not ins:
                        boreholes.append(f_inters[0])
                        len_f_inters = 0  # skips the while loop

            i = 0
            while i < len_f_inters - 1:
                left_offset = [0, 0]
                right_offset = [0, 0]

                # Checks if there is enough distance between this point and another and then will offset the point if
                # there is not enough room
                dls_check = sqrt(
                    (f_inters[i][0] - f_inters[i - 1][0]) * (f_inters[i][0] - f_inters[i - 1][0])
                    + (f_inters[i][1] - f_inters[i - 1][1]) * (f_inters[i][1] - f_inters[i - 1][1])
                )

                drs_check = sqrt(
                    (f_inters[i][0] - f_inters[i + 1][0]) * (f_inters[i][0] - f_inters[i + 1][0])
                    + (f_inters[i][1] - f_inters[i + 1][1]) * (f_inters[i][1] - f_inters[i + 1][1])
                )

                if i > 0 and (dls := dls_check) < x_space:
                    left_offset = [dls * cos(rotate), dls * sin(rotate)]
                elif i < len_f_inters - 1 and (drs := drs_check) < x_space:
                    right_offset = [-drs * cos(rotate), -drs * sin(rotate)]

                process_rows(
                    row,
                    [f_inters[i][0] + left_offset[0], f_inters[i][1] + left_offset[1]],
                    [
                        f_inters[i + 1][0] + right_offset[0],
                        f_inters[i + 1][1] + right_offset[1],
                    ],
                    no_go,
                    x_space,
                    boreholes,
                    rotate=rotate,
                )

                i += 2
        elif len_f_inters == 1:
            ins = False
            for ng_shape in no_go:
                if ng_shape.point_intersect(highest_vert):
                    ins = True
            if not ins:
                if boreholes.size() == 0:
                    boreholes.append(f_inters[0])
                if not (
                    boreholes.index(boreholes.size() - 1)[0] == f_inters[0][0]
                    and boreholes.index(boreholes.size() - 1)[1] == f_inters[0][1]
                ):
                    boreholes.append(f_inters[0])
        else:
            i = 0
            while i < len_f_inters - 1:
                if field.point_intersect(
                    [
                        (f_inters[i][0] + f_inters[i + 1][0]) / 2,
                        (f_inters[i][1] + f_inters[i + 1][1]) / 2,
                    ]
                ):
                    process_rows(
                        row,
                        f_inters[i],
                        f_inters[i + 1],
                        no_go,
                        x_space,
                        boreholes,
                        rotate=rotate,
                    )
                i += 1
        row_point[0] += row_space[0]
        row_point[1] += row_space[1]
    borefield = remove_duplicates(boreholes, duplicate_spacing_ratio * max(x_space, y_space))
    return borefield


def process_rows(row, row_sx, row_ex, no_go, row_space, r_a, rotate, intersection_tolerance=1e-5):
    """
    Function generates a row of the borefield
    *Note: the formatting from the rows can be a little unexpected. Some adjustment
    may be required to correct the formatting. The genBoreHoleConfig function already accounts for this.
    Parameters
    -------------
    :param row:
    :param row_sx:
    :param row_ex:
    :param no_go:
    :param row_space:
    :param r_a:
    :param rotate:
    :param intersection_tolerance:
    """

    if no_go is None:
        distribute(row_sx, row_ex, row_space, r_a)
        return r_a

    # row = [row_sx[0], row_sx[1], row_ex[0], row_ex[1]]

    num_col = int(
        sqrt((row_sx[0] - row_ex[0]) * (row_sx[0] - row_ex[0]) + (row_sx[1] - row_ex[1]) * (row_sx[1] - row_ex[1]))
        // row_space
    )

    inters = [
        point
        for shape in no_go
        for point in shape.line_intersect(row, rotate=rotate, intersection_tolerance=intersection_tolerance)
    ]
    inters = sort_intersections(inters, rotate)
    num_inters = len(inters)

    if num_inters > 1:  # noqa: SIM102
        if less_than(
            inters[0],
            row_sx,
            rotate=rotate,
            intersection_tolerance=intersection_tolerance,
        ) and less_than(
            row_ex,
            inters[len(inters) - 1],
            rotate=rotate,
            intersection_tolerance=intersection_tolerance,
        ):
            inside = False
            for _ in inters:
                if less_than(
                    row_sx,
                    inters[0],
                    rotate=rotate,
                    intersection_tolerance=intersection_tolerance,
                ) and less_than(
                    inters[0],
                    row_ex,
                    rotate=rotate,
                    intersection_tolerance=intersection_tolerance,
                ):
                    inside = True
            if not inside:
                point_in = False
                for ng_shape in no_go:
                    if ng_shape.point_intersect([(row_ex[0] + row_sx[0]) / 2, (row_ex[1] + row_sx[1]) / 2]):
                        point_in = True
                if point_in:
                    return []
    inters = np.array(inters)
    indices = []
    for j in range(num_inters):
        less_than_1 = less_than(row_ex, inters[j], rotate=rotate, intersection_tolerance=intersection_tolerance)
        less_than_2 = less_than(inters[j], row_sx, rotate=rotate, intersection_tolerance=intersection_tolerance)
        if not (less_than_1 or less_than_2):
            indices.append(j)
    inters = inters[indices]
    num_inters = len(inters)
    for i in range(num_inters - 1):
        space = sqrt(
            (inters[i + 1][0] - inters[i][0]) * (inters[i + 1][0] - inters[i][0])
            + (inters[i + 1][1] - inters[i][1]) * (inters[i + 1][1] - inters[i][1])
        )
        if space < row_space:
            i_none = False
            for shape in no_go:
                if shape.point_intersect(
                    [
                        (inters[i + 1][0] + inters[i][0]) / 2,
                        (inters[i + 1][1] + inters[i][1]) / 2,
                    ]
                ):
                    i_none = True
            if i_none:
                d = (row_space - space) / 2
                inters[i + 1][0] += d * cos(rotate)
                inters[i + 1][1] += d * sin(rotate)
                inters[i][0] -= d * cos(rotate)
                inters[i][1] -= d * sin(rotate)
    if num_col < 1:
        ins = False
        for shape in no_go:
            if shape.point_intersect([(row_ex[0] + row_sx[0]) / 2, (row_ex[1] + row_sx[1]) / 2]):
                ins = True
        if not ins:
            if r_a.size() == 0 or not (
                r_a.index(r_a.size() - 1)[0] == (row_ex[0] + row_sx[0]) / 2
                and r_a.index(r_a.size() - 1)[1] == (row_ex[1] + row_sx[1]) / 2
            ):
                r_a.append([(row_ex[0] + row_sx[0]) / 2, (row_ex[1] + row_sx[1]) / 2])
            return r_a
    elif num_inters == 0:
        if not_inside(row_sx, no_go) and not_inside(row_ex, no_go):
            distribute(row_sx, row_ex, row_space, r_a)
    elif num_inters == 2:  # noqa: PLR2004
        distribute(row_sx, inters[0], row_space, r_a)
        distribute(inters[1], row_ex, row_space, r_a)
    elif num_inters == 1:
        ins = False
        for shape in no_go:
            if shape.point_intersect([(inters[0][0] + row_sx[0]) / 2, (inters[0][1] + row_sx[1]) / 2]):
                ins = True
        if not ins:
            distribute(row_sx, inters[0], row_space, r_a)
        else:
            distribute(inters[0], row_ex, row_space, r_a)
    elif num_inters % 2 == 0:
        i = 0
        while i < num_inters:
            if i == 0:
                distribute(row_sx, inters[0], row_space, r_a)
                i = 1
                continue
            elif i == num_inters - 1:
                distribute(inters[num_inters - 1], row_ex, row_space, r_a)
            else:
                distribute(inters[i], inters[i + 1], row_space, r_a)
            i += 2
    else:
        ins = False
        for shape in no_go:
            if shape.point_intersect([(inters[0][0] + row_sx[0]) / 2, (inters[0][1] + row_sx[1]) / 2]):
                ins = True
        if not ins:
            i = 0
            while i < num_inters:
                if i == 0:
                    distribute(row_sx, inters[0], row_space, r_a)
                    i = 1
                    continue
                elif i == num_inters - 1:
                    i += 2
                    continue
                else:
                    distribute(inters[i], inters[i + 1], row_space, r_a)
                i += 2
        else:
            i = 0
            while i < num_inters:
                if i == 0:
                    distribute(inters[0], inters[1], row_space, r_a)
                    i = 2
                    continue
                elif i == num_inters - 1:
                    distribute(inters[i], row_ex, row_space, r_a)
                    i += 2
                    continue
                else:
                    distribute(inters[i], inters[i + 1], row_space, r_a)
                i += 2

    return r_a


def not_inside(p, ngs):
    inside = False
    for ng in ngs:
        if ng.point_intersect(p):
            inside = True
    return not inside


def less_than(p1, p2, rotate=0, intersection_tolerance=1e-5):
    x1, y1 = p1
    x2, y2 = p2
    dx = x2 - x1
    dy = y2 - y1

    if abs(dx) < intersection_tolerance:
        dx_sign = 0
    elif dx > 0:
        dx_sign = 1
    else:
        dx_sign = -1

    if abs(dy) < intersection_tolerance:
        dy_sign = 0
    elif dy > 0:
        dy_sign = 1
    else:
        dy_sign = -1

    if rotate >= 0:
        if dx_sign == 0:
            return dy_sign == 1
        elif dy_sign in (0, dx_sign):
            return dx_sign == 1
        else:
            raise ValueError("Slope between points does not match field orientation.")
    elif dx_sign == 0:
        if dy_sign == 1:
            return False
        else:
            return False
    elif dy_sign == 0 or dx_sign != dy_sign:
        return dx_sign == 1
    else:
        raise ValueError("Slope between points does not match field orientation.")


@lru_cache(maxsize=1024)
def evenly_spaced_points(number_of_points, starting_val=0.0, ending_val=1.0):
    return np.linspace(starting_val, ending_val, num=number_of_points, endpoint=True)


def distribute(x1, x2, spacing, r):
    """
      Function generates a series of boreholes between x1 and x2
    Parameters
    -------------
    :param x1: float
        left x value
    :param x2: float
        right x value
    :param spacing: float
        spacing between columns
    :param r: [[float]]
        existing array of points
    :param rotate:
    :return:
    """
    # This is the squared distance between the two row edges.
    dx = x2[0] - x1[0]
    dy = x2[1] - x1[1]
    d = sqrt(dx * dx + dy * dy)

    # Check if there is only room for one point along the row
    if d < spacing:
        # If we have currently not generated any coordinates, or we have not added a point at the same location
        # previously, then add at the average between the two row edges.
        if r.size() == 0 or not (
            r.index(r.size() - 1)[0] == (x1[0] + x2[0]) / 2 and r.index(r.size() - 1)[1] == (x1[1] + x2[1]) / 2
        ):
            r.append([(x1[0] + x2[0]) / 2, (x1[1] + x2[1]) / 2])
        return r

    # Determine the number of boreholes placed along the row.
    act_num_col = int(d // spacing) + 1

    # Determine the x and y values along the row
    spacing_multipliers = evenly_spaced_points(act_num_col)
    x_vals = x1[0] + dx * spacing_multipliers
    y_vals = x1[1] + dy * spacing_multipliers
    r.extend(zip(x_vals, y_vals))
    return r



from abc import abstractmethod
from dataclasses import dataclass, field

from pygfunction.boreholes import Borehole

from ghedesigner.enums import DesignGeomType, FlowConfigType, TimestepType
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.ghe.search.bisection_1d import Bisection1D
from ghedesigner.ghe.search.bisection_2d import Bisection2D
from ghedesigner.ghe.search.bisection_zd import BisectionZD
from ghedesigner.ghe.search.rowwise import RowWiseModifiedBisectionSearch
from ghedesigner.media import Fluid, Grout, Soil

AnyBisectionType = Bisection1D | Bisection2D | BisectionZD | RowWiseModifiedBisectionSearch


@dataclass(init=False)
class GeometricConstraints:
    type: DesignGeomType = field(default=DesignGeomType.NONE, init=False, repr=False)

    @abstractmethod
    def to_input(self) -> dict:
        pass


@dataclass
class DesignBase:
    def __init__(
        self,
        v_flow: float,
        borehole: Borehole,
        fluid: Fluid,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        start_month: int,
        end_month: int,
        max_eft: float,
        min_eft: float,
        max_height: float,
        min_height: float,
        continue_if_design_unmet: bool,
        max_boreholes: int | None,
        geometric_constraints: GeometricConstraints,
        hourly_extraction_ground_loads: list,
        method: TimestepType,
        flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
        load_years=None,
    ) -> None:
        if load_years is None:
            load_years = [2019]
        self.load_years = load_years
        self.v_flow = v_flow  # volumetric flow rate, m3/s
        self.borehole = borehole
        self.fluid = fluid  # a fluid object
        self.pipe = pipe
        self.grout = grout
        self.soil = soil
        self.geometric_constraints = geometric_constraints
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.method = method
        self.flow_type = flow_type
        self.start_month = start_month
        self.end_month = end_month
        self.max_EFT_allowable = max_eft
        self.min_EFT_allowable = min_eft
        self.max_height = max_height
        self.min_height = min_height
        self.continue_if_design_unmet = continue_if_design_unmet
        self.max_boreholes = max_boreholes
        if self.method == "hourly":
            msg = (
                "Note: It is not recommended to perform a field selection \n",
                "with the hourly simulation due to computation time. If \n",
                "the goal is to validate the selected field with the \n",
                "hourly simulation, the better solution is to utilize the \n",
                "hybrid simulation to automatically select the field. Then \n",
                "perform a sizing routine on the selected GHE with the \n",
                "hourly simulation.",
            )
            for element in msg:
                print(element)
            print("\n")

    @abstractmethod
    def find_design(self, disp=False) -> AnyBisectionType:
        pass

    def to_input(self) -> dict:
        return {"flow_rate": self.v_flow, "flow_type": self.flow_type.name}



from enum import Enum, auto


class PipeType(Enum):
    COAXIAL = "COAXIAL"
    DOUBLEUTUBEPARALLEL = "DOUBLEUTUBEPARALLEL"
    DOUBLEUTUBESERIES = "DOUBLEUTUBESERIES"
    SINGLEUTUBE = "SINGLEUTUBE"


class DoubleUTubeConnType(Enum):
    PARALLEL = auto()
    SERIES = auto()


class TimestepType(Enum):
    HOURLY = auto()
    HYBRID = auto()


class DesignGeomType(Enum):
    BIRECTANGLE = auto()
    BIRECTANGLECONSTRAINED = auto()
    BIZONEDRECTANGLE = auto()
    NEARSQUARE = auto()
    RECTANGLE = auto()
    ROWWISE = auto()
    NONE = auto()  # TODO: Check this won't break anything


class FlowConfigType(Enum):
    BOREHOLE = "BOREHOLE"
    SYSTEM = "SYSTEM"


class FluidType(Enum):
    ETHYLALCOHOL = auto()
    ETHYLENEGLYCOL = auto()
    METHYLALCOHOL = auto()
    PROPYLENEGLYCOL = auto()
    WATER = auto()



# This is just a very thin wrapper over the pygfunction borehole class
# TODO: Think about how this class plays with the other actual configured borehole classes, could just be a rename

from typing import Any

from pygfunction.boreholes import Borehole as PyG_Borehole


class Borehole(PyG_Borehole):
    def __init__(
        self,
        burial_depth: float,
        borehole_radius: float,
        borehole_height: float = 100,  # TODO: Not sure this value is ever used, but we set it to 100 everywhere
        tilt: float = 0.0,
        orientation: float = 0.0,
        x: Any = 0.0,
        y: Any = 0.0,
    ):
        super().__init__(
            H=borehole_height, D=burial_depth, r_b=borehole_radius, x=x, y=y, tilt=tilt, orientation=orientation
        )



from math import ceil

import numpy as np
from pygfunction.boreholes import Borehole

from ghedesigner.enums import FlowConfigType, TimestepType
from ghedesigner.ghe.gfunction import calc_g_func_for_multiple_lengths
from ghedesigner.ghe.ground_heat_exchangers import GHE
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.media import Fluid, Grout, Soil
from ghedesigner.utilities import borehole_spacing, check_bracket, eskilson_log_times, sign


class Bisection1D:
    def __init__(
        self,
        coordinates_domain: list,
        field_descriptors: list,
        v_flow: float,
        borehole: Borehole,
        fluid: Fluid,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        max_boreholes: int | None,
        min_height: float,
        max_height: float,
        continue_if_design_unmet: bool,
        start_month: int,
        end_month: int,
        min_eft: float,
        max_eft: float,
        hourly_extraction_ground_loads: list,
        method: TimestepType,
        flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
        max_iter=15,
        disp=False,
        search=True,
        field_type="N/A",
        load_years=None,
    ) -> None:
        # Take the lowest part of the coordinates domain to be used for the
        # initial setup
        if load_years is None:
            load_years = [2019]
        self.load_years = load_years
        self.searchTracker: list[list] = []
        coordinates = coordinates_domain[0]
        current_field = field_descriptors[0]
        self.field_type = field_type
        # Flow rate tracking
        self.v_flow = v_flow
        self.flow_type = flow_type
        v_flow_system, m_flow_borehole = self.retrieve_flow(coordinates, fluid.rho)
        self.method = method
        self.min_height = min_height
        self.max_height = max_height
        self.start_month = start_month
        self.end_month = end_month
        self.max_boreholes = max_boreholes
        self.max_eft = max_eft
        self.min_eft = min_eft
        self.continue_if_design_unmet = continue_if_design_unmet
        self.log_time = eskilson_log_times()
        self.bhe_type = pipe.type
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.coordinates_domain = coordinates_domain
        self.fieldDescriptors = field_descriptors
        self.max_iter = max_iter
        self.disp = disp

        b = borehole_spacing(borehole, coordinates)

        # Calculate a g-function for uniform inlet fluid temperature with
        # 8 unequal segments using the equivalent solver
        g_function = calc_g_func_for_multiple_lengths(
            b,
            [borehole.H],
            borehole.r_b,
            borehole.D,
            m_flow_borehole,
            self.bhe_type,
            self.log_time,
            coordinates,
            fluid,
            pipe,
            grout,
            soil,
        )

        # Initialize the GHE object
        self.ghe = GHE(
            v_flow_system,
            b,
            pipe.type,
            fluid,
            borehole,
            pipe,
            grout,
            soil,
            g_function,
            self.start_month,
            self.end_month,
            hourly_extraction_ground_loads,
            field_specifier=current_field,
            field_type=field_type,
        )

        self.calculated_temperatures: dict[int, np.float64] = {}

        if search:
            self.selection_key, self.selected_coordinates = self.search()

    def retrieve_flow(self, coordinates, rho):
        if self.flow_type == FlowConfigType.BOREHOLE:
            v_flow_system = self.v_flow * len(coordinates)
            # Total fluid mass flow rate per borehole (kg/s)
            m_flow_borehole = self.v_flow / 1000.0 * rho
        elif self.flow_type == FlowConfigType.SYSTEM:
            v_flow_system = self.v_flow
            v_flow_borehole = self.v_flow / len(coordinates)
            m_flow_borehole = v_flow_borehole / 1000.0 * rho
        else:
            raise ValueError("The flow argument should be either `borehole` or `system`.")
        return v_flow_system, m_flow_borehole

    def initialize_ghe(self, coordinates, h, field_specifier="N/A"):
        v_flow_system, m_flow_borehole = self.retrieve_flow(coordinates, self.ghe.bhe.fluid.rho)

        self.ghe.bhe.borehole.H = h
        borehole = self.ghe.bhe.borehole
        fluid = self.ghe.bhe.fluid
        pipe = self.ghe.bhe.pipe
        grout = self.ghe.bhe.grout
        soil = self.ghe.bhe.soil

        b = borehole_spacing(borehole, coordinates)

        # Calculate a g-function for uniform inlet fluid temperature with
        # 8 unequal segments using the equivalent solver
        g_function = calc_g_func_for_multiple_lengths(
            b,
            [borehole.H],
            borehole.r_b,
            borehole.D,
            m_flow_borehole,
            self.bhe_type,
            self.log_time,
            coordinates,
            fluid,
            pipe,
            grout,
            soil,
        )

        # Initialize the GHE object
        self.ghe = GHE(
            v_flow_system,
            b,
            self.bhe_type,
            fluid,
            borehole,
            pipe,
            grout,
            soil,
            g_function,
            self.start_month,
            self.end_month,
            self.hourly_extraction_ground_loads,
            field_type=self.field_type,
            field_specifier=field_specifier,
        )

    def calculate_excess(self, coordinates, h, field_specifier="N/A"):
        self.initialize_ghe(coordinates, h, field_specifier=field_specifier)
        # Simulate after computing just one g-function
        max_hp_eft, min_hp_eft = self.ghe.simulate(method=self.method)
        t_excess = self.ghe.cost(max_hp_eft, min_hp_eft, self.max_eft, self.min_eft)
        self.searchTracker.append([field_specifier, t_excess, max_hp_eft, min_hp_eft])

        return t_excess

    def search(self):
        x_l_idx = 0

        # find upper bound that respects max_boreholes
        if self.max_boreholes is not None:
            num_coordinates_in_each = [len(x) for x in self.coordinates_domain]
            x_r_idx = [idx for idx, x in enumerate(num_coordinates_in_each) if x < self.max_boreholes][-1]
        else:
            x_r_idx = len(self.coordinates_domain) - 1

        if self.disp:
            print("Do some initial checks before searching.")
        # Get the lowest possible excess temperature from minimum height at the
        # smallest location in the domain
        t_0_lower = self.calculate_excess(
            self.coordinates_domain[x_l_idx],
            self.min_height,
            field_specifier=self.fieldDescriptors[x_l_idx],
        )
        t_0_upper = self.calculate_excess(
            self.coordinates_domain[x_l_idx],
            self.max_height,
            field_specifier=self.fieldDescriptors[x_l_idx],
        )
        t_m1 = self.calculate_excess(
            self.coordinates_domain[x_r_idx],
            self.max_height,
            field_specifier=self.fieldDescriptors[x_r_idx],
        )

        self.calculated_temperatures[x_l_idx] = t_0_upper
        self.calculated_temperatures[x_r_idx] = t_m1

        if check_bracket(sign(t_0_lower), sign(t_0_upper)):
            if self.disp:
                print("Size between min and max of lower bound in domain.")
            self.initialize_ghe(self.coordinates_domain[x_l_idx], self.max_height)
            return x_l_idx, self.coordinates_domain[x_l_idx]
        elif check_bracket(sign(t_0_upper), sign(t_m1)):
            if self.disp:
                print("Perform the integer bisection search routine.")
        elif t_0_lower < 0.0:
            condition_msg = (
                "The optimal design requires fewer or shorter boreholes \n"
                "than what is possible based on the current design parameters."
            )
            print(condition_msg)
            if self.continue_if_design_unmet:
                print("Smallest available configuration selected.")
                selection_key = x_l_idx
                self.initialize_ghe(
                    self.coordinates_domain[selection_key],
                    self.min_height,
                    self.fieldDescriptors[selection_key],
                )
                return selection_key, self.coordinates_domain[selection_key]
            else:
                raise ValueError("Search failed.")
        elif t_m1 > 0.0:
            condition_msg = (
                "The optimal design requires more or deeper boreholes \n"
                "than what is possible based on the current design parameters. \n"
                "Consider increasing the available land area, \n"
                "increasing the maximum borehole depth, \n"
                "or decreasing the maximum borehole spacing."
            )
            print(condition_msg)
            if self.continue_if_design_unmet:
                print("Largest available configuration selected.")
                selection_key = x_r_idx
                self.initialize_ghe(
                    self.coordinates_domain[selection_key],
                    self.max_height,
                    self.fieldDescriptors[selection_key],
                )
                return selection_key, self.coordinates_domain[selection_key]
            else:
                raise ValueError("Search failed.")
        else:
            # if we've gotten here, everything should be good for the bisection search.
            # can add catches for other cases here if they pop up.
            pass
        if self.disp:
            print("Beginning bisection search...")

        x_l_sign = sign(t_0_upper)

        i = 0

        while i < self.max_iter:
            c_idx = ceil((x_l_idx + x_r_idx) / 2)
            # if the solution is no longer making progress break the while
            # if c_idx == x_l_idx or c_idx == x_r_idx:
            if c_idx in (x_l_idx, x_r_idx):
                break

            c_t_excess = self.calculate_excess(
                self.coordinates_domain[c_idx],
                self.max_height,
                field_specifier=self.fieldDescriptors[c_idx],
            )

            self.calculated_temperatures[c_idx] = c_t_excess
            c_sign = sign(c_t_excess)

            if c_sign == x_l_sign:
                x_l_idx = c_idx
            else:
                x_r_idx = c_idx

            i += 1

        coordinates = self.coordinates_domain[i]

        self.calculate_excess(coordinates, self.max_height, self.fieldDescriptors[i])
        # Make sure the field being returned pertains to the index which is the
        # closest to 0 but also negative (the maximum of all 0 or negative
        # excess temperatures)
        keys = list(self.calculated_temperatures.keys())
        values = list(self.calculated_temperatures.values())

        # theoretically, the biggest negative value should be the field that is just undersized
        negative_excess_values = [v for v in values if v <= 0.0]
        excess_of_interest = max(negative_excess_values)

        # but some conditions don't yield this result
        # adding a check here to ensure we pick the smallest field with
        # negative excess temperature
        num_bh = [len(self.coordinates_domain[x]) for x in keys]
        sorted_num_bh, sorted_values = (list(t) for t in zip(*sorted(zip(num_bh, values))))
        for _, val in zip(sorted_num_bh, sorted_values):
            if val < 0:
                if excess_of_interest != val:
                    print(
                        "Loads resulted in odd behavior requiring the selected field configuration \n"
                        "to be reset to the smallest field with negative excess temperature. \n"
                        "Please forward the inputs to the developers for investigation."
                    )
                excess_of_interest = val
                break

        idx = values.index(excess_of_interest)
        selection_key = keys[idx]
        self.initialize_ghe(
            self.coordinates_domain[selection_key], self.max_height, self.fieldDescriptors[selection_key]
        )
        return selection_key, self.coordinates_domain[selection_key]



import csv
from json import dumps
from pathlib import Path
from typing import Any

from numpy import ndarray

from ghedesigner.enums import TimestepType
from ghedesigner.ghe.design.base import AnyBisectionType
from ghedesigner.ghe.ground_heat_exchangers import GHE
from ghedesigner.output.converters import ghe_time_convert, hours_to_month
from ghedesigner.output.json_serializer import JsonSerializer
from ghedesigner.output.text_serializer import TextSerializer
from ghedesigner.utilities import write_flat_dict_to_csv, write_json


class OutputManager:
    """
    Orchestrates writing of all GHE design outputs:
      - text summary
      - multiple CSVs
      - JSON summary
    """

    def __init__(
        self,
        project_name: str,
        notes: str,
        author: str,
        model_name: str,
        allocated_width: int = 100,
    ) -> None:
        self.project_name = project_name
        self.notes = notes
        self.author = author
        self.model_name = model_name
        self.allocated_width = allocated_width

        self.design: AnyBisectionType | None = None
        self.time: float = 0.0
        self.load_method: TimestepType = TimestepType.HYBRID

    @staticmethod
    def just_write_g_function(
        output_directory: Path,
        log_time: ndarray,
        g_values: ndarray,
        g_bhw_values: ndarray,
    ) -> None:
        output_directory.mkdir(parents=True, exist_ok=True)
        summary = {
            "log_time": log_time.tolist(),
            "g_values": g_values.tolist(),
            "g_bhw_values": g_bhw_values.tolist(),
        }
        write_json(output_directory / "SimulationSummary.json", summary)
        write_flat_dict_to_csv(output_directory / "Gfunction.csv", summary)

    def set_design_data(
        self,
        design: AnyBisectionType,
        time: float,
        load_method: TimestepType,
    ) -> None:
        """Store design result and runtime for later writing."""
        self.design = design
        self.time = time
        self.load_method = load_method

    def write_all_output_files(
        self,
        output_directory: Path,
        file_suffix: str = "",
    ) -> None:
        """Write .txt, .csv and .json outputs for a full simulation."""
        if self.design is None or self.design.ghe is None:
            raise ValueError("Design data has not been set")

        ghe = self.design.ghe
        tracker = self.design.searchTracker

        output_directory.mkdir(parents=True, exist_ok=True)

        # Text summary
        txt = TextSerializer.summary_text(
            self.allocated_width,
            self.project_name,
            self.model_name,
            self.notes,
            self.author,
            self.time,
            ghe,
            tracker,
        )
        (output_directory / f"SimulationSummary{file_suffix}.txt").write_text(txt)

        # CSVs
        with open(output_directory / f"TimeDependentValues{file_suffix}.csv", "w", newline="") as f:
            csv.writer(f).writerows(self._get_loading_data(ghe))

        with open(output_directory / f"BoreFieldData{file_suffix}.csv", "w", newline="") as f:
            csv.writer(f).writerows(self._get_borehole_location_data(ghe))

        with open(output_directory / f"Loadings{file_suffix}.csv", "w", newline="") as f:
            csv.writer(f).writerows(self._get_hourly_loading_data(ghe))

        with open(output_directory / f"Gfunction{file_suffix}.csv", "w", newline="") as f:
            csv.writer(f).writerows(self._get_g_function_data(ghe))

        # JSON summary
        obj = JsonSerializer.summary_object(
            ghe,
            tracker,
            self.time,
            self.project_name,
            self.notes,
            self.author,
            self.model_name,
            self.load_method,
        )

        with open(output_directory / f"SimulationSummary{file_suffix}.json", "w", newline="") as f:
            f.write(dumps(obj, indent=2))

    def write_presized_output_files(
        self,
        output_directory: Path,
        ghe: GHE,
        file_suffix: str = "",
    ) -> None:
        """Write minimal outputs for a presized design."""
        output_directory.mkdir(parents=True, exist_ok=True)

        txt = TextSerializer.summary_text(
            self.allocated_width,
            self.project_name,
            self.model_name,
            self.notes,
            self.author,
            self.time,
            ghe,
            "none",
        )
        (output_directory / f"SimulationSummary{file_suffix}.txt").write_text(txt)

        with open(output_directory / f"BoreFieldData{file_suffix}.csv", "w", newline="") as f:
            csv.writer(f).writerows(self._get_borehole_location_data(ghe))

        with open(output_directory / f"Gfunction{file_suffix}.csv", "w", newline="") as f:
            csv.writer(f).writerows(self._get_g_function_data(ghe))

        # This is commented out because it was writing dummy data
        # write_json(output_directory / f"SimulationSummary{file_suffix}.json", {
        #     "ghe_system": {
        #         "number_of_boreholes": 1,
        #         "active_borehole_length": {"value": 1},
        #     }
        # })

    # Internal data extraction for CSVs
    def _get_loading_data(self, ghe: GHE) -> list[list[Any]]:
        times = ghe.times
        d_tb = ghe.dTb
        hp_eft = ghe.hp_eft
        loading = ghe.loading
        denom = ghe.bhe.borehole.H * ghe.nbh
        ugt = ghe.bhe.soil.ugt

        rows: list[list[Any]] = [
            [
                "Time (hr)",
                "Time (month)",
                "Q (Rejection) (W) (before time)",
                "Q (Rejection) (W/m) (before time)",
                "Tb (C)",
                "GHE ExFT (C)",
            ]
        ]

        n = len(times)
        for i, t in enumerate(times):
            month = hours_to_month(t)

            # "Before time" row (uses current loading; Tb/EFT from previous index)
            q_before = loading[i] if loading is not None and i > 1 else 0
            rows.append(
                [
                    t,
                    month,
                    q_before,
                    (q_before / denom) if i > 1 else 0,
                    # TODO The next two lines wrap to the last element when i==0, it's not clear if that's intentional
                    ugt + d_tb[i - 1],
                    hp_eft[i - 1],
                ]
            )

            # "After time" row (uses next loading; current Tb/EFT)
            q_after = loading[i + 1] if loading is not None and (i + 1) < n else 0
            rows.append(
                [
                    t,
                    month,
                    q_after,
                    (q_after / denom) if q_after else 0,
                    ugt + d_tb[i],
                    hp_eft[i],
                ]
            )

        return rows

    def _get_borehole_location_data(self, ghe: GHE) -> list[list[Any]]:
        return [["x", "y"]] + [[x, y] for x, y in ghe.gFunction.bore_locations]

    def _get_hourly_loading_data(self, ghe: GHE) -> list[list[Any]]:
        rows: list[list[Any]] = [["Month", "Day", "Hour", "Time (Hours)", "Loading (W) (Extraction)"]]
        for hr, load in enumerate(ghe.hourly_extraction_ground_loads):
            m, d, h = ghe_time_convert(hr)
            rows.append([m, d, h, hr, load])
        return rows

    def _get_g_function_data(self, ghe: GHE) -> list[list[Any]]:
        title = f"H: {ghe.bhe.borehole.H:0.2f} m"
        gf_adj, gf_bhw = ghe.grab_g_function(ghe.b_spacing / ghe.bhe.borehole.H)
        header = ["ln(t/ts)", title, f"{title} bhw"]
        return [header] + [[x, y, z] for x, y, z in zip(gf_adj.x, gf_adj.y, gf_bhw.y)]



import warnings
from typing import Any


def create_title(width: int, title: str, filler_symbol: str = " ") -> str:
    """Centered title line, e.g. `----- Results -----`"""
    return f"{f' {title} ':{filler_symbol}^{width}s}\n"


def create_row(
    width: int,
    row_data: list[Any],
    data_formats: list[str],
    centering: str = ">",
) -> str:
    """
    Format a single row of data into fixed-width columns.
    data_formats: list of format specifiers like ".2f" or "s", one per column.
    """
    out = ""
    n_cols = len(row_data)
    col_w = width // n_cols
    leftover = width % n_cols
    for datum, fmt in zip(row_data, data_formats):
        w = col_w + (1 if leftover > 0 else 0)
        leftover -= 1 if leftover > 0 else 0
        try:
            out += f"{datum:{centering}{w}{fmt}}"
        except Exception as e:
            print(f"Row formatting error ({fmt}):", e)
            raise
    return f"{out}\n"


def create_table(
    title: str,
    col_titles: list[list[str]],
    rows: list[list[Any]],
    width: int,
    col_formats: list[str],
    filler_symbol: str = " ",
    centering: str = ">",
) -> str:
    """
    Build a table with a header title, column headers, dashed lines, and data rows.
    col_titles: list of header rows, each a list of strings.
    """
    txt = create_title(width, title, filler_symbol)
    txt += create_line(width)
    header_fmt = ["s"] * len(col_titles[0])
    for header_row in col_titles:
        txt += create_row(width, header_row, header_fmt, centering="^")
    txt += create_line(width)
    for row in rows:
        txt += create_row(width, row, col_formats, centering=centering)
    txt += create_line(width)
    return txt


def d_row(
    width: int,
    label: str,
    value: Any,
    fmt: str,
    n_tabs: int = 0,
) -> str:
    """
    A simple two-column row: left label, right value, padded in between.
    n_tabs inserts 4*n_tabs spaces before label.
    """
    tab_w = 4
    lead = " " * (n_tabs * tab_w)
    left = f"{lead}{label}"
    right = f"{value:{fmt}}"
    pad = width - len(left) - len(right)
    if pad < 0:
        warnings.warn("Formatting width too small for d_row; truncating")
        pad = 4
    return f"{left}{' ' * pad}{right}\n"


def create_line(width: int, character: str = "*") -> str:
    """A full-width line of a single repeat character."""
    return f"{character * width}\n"



from pygfunction.boreholes import Borehole

from ghedesigner.enums import DoubleUTubeConnType, PipeType
from ghedesigner.ghe.boreholes.coaxial_borehole import CoaxialPipe
from ghedesigner.ghe.boreholes.multi_u_borehole import MultipleUTube
from ghedesigner.ghe.boreholes.single_u_borehole import SingleUTube
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.media import Fluid, Grout, Soil


def get_bhe_object(
    bhe_type: PipeType,
    m_flow_borehole: float,
    fluid: Fluid,
    borehole: Borehole,
    pipe: Pipe,
    grout: Grout,
    soil: Soil,
):
    if bhe_type == PipeType.SINGLEUTUBE:
        return SingleUTube(m_flow_borehole, fluid, borehole, pipe, grout, soil)
    elif bhe_type == PipeType.DOUBLEUTUBEPARALLEL:
        return MultipleUTube(m_flow_borehole, fluid, borehole, pipe, grout, soil, config=DoubleUTubeConnType.PARALLEL)
    elif bhe_type == PipeType.DOUBLEUTUBESERIES:
        return MultipleUTube(m_flow_borehole, fluid, borehole, pipe, grout, soil, config=DoubleUTubeConnType.SERIES)
    elif bhe_type == PipeType.COAXIAL:
        return CoaxialPipe(m_flow_borehole, fluid, borehole, pipe, grout, soil)
    else:
        raise TypeError("BHE type not implemented")



from dataclasses import asdict, dataclass, field

from pygfunction.boreholes import Borehole

from ghedesigner.enums import DesignGeomType, FlowConfigType, TimestepType
from ghedesigner.ghe.design.base import DesignBase, GeometricConstraints
from ghedesigner.ghe.domains import rectangular
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.ghe.search.bisection_1d import Bisection1D
from ghedesigner.media import Fluid, Grout, Soil


@dataclass
class GeometricConstraintsRectangle(GeometricConstraints):
    """
    Geometric constraints for rectangular design algorithm
    """

    length: float
    width: float
    b_min: float
    b_max: float
    type: DesignGeomType = field(default=DesignGeomType.RECTANGLE, init=False, repr=False)

    def to_input(self) -> dict:
        return {
            **asdict(self, dict_factory=lambda d: {k: v for k, v in d if k != "type"}),
            "method": self.type.name,
        }


class DesignRectangle(DesignBase):
    def __init__(
        self,
        v_flow: float,
        borehole: Borehole,
        fluid: Fluid,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        start_month: int,
        end_month: int,
        max_eft: float,
        min_eft: float,
        max_height: float,
        min_height: float,
        continue_if_design_unmet: bool,
        max_boreholes: int | None,
        geometric_constraints: GeometricConstraintsRectangle,
        hourly_extraction_ground_loads: list,
        method: TimestepType,
        flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
        load_years=None,
    ) -> None:
        super().__init__(
            v_flow,
            borehole,
            fluid,
            pipe,
            grout,
            soil,
            start_month,
            end_month,
            max_eft,
            min_eft,
            max_height,
            min_height,
            continue_if_design_unmet,
            max_boreholes,
            geometric_constraints,
            hourly_extraction_ground_loads,
            method,
            flow_type,
            load_years,
        )
        self.geometric_constraints = geometric_constraints
        self.coordinates_domain, self.fieldDescriptors = rectangular(
            self.geometric_constraints.length,
            self.geometric_constraints.width,
            self.geometric_constraints.b_min,
            self.geometric_constraints.b_max,
        )

    def find_design(self, disp=False) -> Bisection1D:
        if disp:
            title = "Find rectangle..."
            print(title + "\n" + len(title) * "=")
        return Bisection1D(
            self.coordinates_domain,
            self.fieldDescriptors,
            self.v_flow,
            self.borehole,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.max_boreholes,
            self.min_height,
            self.max_height,
            self.continue_if_design_unmet,
            self.start_month,
            self.end_month,
            self.min_EFT_allowable,
            self.max_EFT_allowable,
            self.hourly_extraction_ground_loads,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="rectangle",
            load_years=self.load_years,
        )



from enum import IntEnum, auto
from math import exp, log, pi, sqrt
from typing import cast

import numpy as np
from bhr.borehole import Borehole as BHRBorehole
from pygfunction.boreholes import Borehole
from scipy.interpolate import interp1d
from scipy.linalg.lapack import dgtsv

from ghedesigner.constants import SEC_IN_HR, TWO_PI
from ghedesigner.ghe.boreholes.base import GHEDesignerBoreholeBase
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.media import Fluid, Grout, Soil


class CellProps(IntEnum):
    R_IN = 0
    R_CENTER = auto()
    R_OUT = auto()
    K = auto()
    RHO_CP = auto()
    TEMP = auto()
    VOL = auto()


class SingleUTube(GHEDesignerBoreholeBase):
    def __init__(
        self,
        m_flow_borehole: float,
        fluid: Fluid,
        borehole: Borehole,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
    ) -> None:
        GHEDesignerBoreholeBase.__init__(self, m_flow_borehole, fluid, borehole, pipe, grout, soil)
        self.R_fp = 0.0
        self.fluid = fluid
        self.m_flow_borehole = m_flow_borehole
        self.borehole = borehole
        self.pipe = pipe
        self.soil = soil
        self.grout = grout

        self.bhr_borehole = BHRBorehole()

        # Ensure r_in and r_out are floats rather than list[float]
        if not isinstance(pipe.r_out, float) or not isinstance(pipe.r_in, float):
            raise TypeError("pipe r_in and r_out must be floats")

        self.bhr_borehole.init_single_u_borehole(
            borehole_diameter=borehole.r_b * 2,
            pipe_outer_diameter=(2.0 * pipe.r_out),
            pipe_dimension_ratio=(2.0 * pipe.r_out) / (pipe.r_out - pipe.r_in),
            length=borehole.H,
            shank_space=(pipe.s / 2.0 + pipe.r_out),
            pipe_conductivity=pipe.k,
            grout_conductivity=grout.k,
            soil_conductivity=soil.k,
            fluid_type=self.fluid.name,
            fluid_concentration=self.fluid.concentration_percent / 100,
            boundary_condition="UNIFORM_BOREHOLE_WALL_TEMP",
        )

        # radial numerical initialization
        # "The one dimensional model has a fluid core, an equivalent convective
        # resistance layer, a tube layer, a grout layer and is surrounded by the
        # ground."

        # cell numbers
        self.num_fluid_cells = 3
        self.num_conv_cells = 1
        self.num_pipe_cells = 4
        self.num_grout_cells = 27
        self.num_soil_cells = 500

        self.num_cells = sum(
            (self.num_fluid_cells, self.num_conv_cells, self.num_pipe_cells, self.num_grout_cells, self.num_soil_cells)
        )
        self.bh_wall_idx = sum((self.num_fluid_cells, self.num_conv_cells, self.num_pipe_cells, self.num_grout_cells))

        # Geometry and grid procedure

        # far-field radius is set to 10m
        self.r_far_field = 10

        # borehole radius is set to the actual radius of the borehole
        self.r_borehole = self.borehole.r_b

        # outer tube radius is set to sqrt(2) * r_p_o, tube region has 4 cells
        r_out = cast(float, self.pipe.r_out)
        r_in = cast(float, self.pipe.r_in)

        self.r_out_tube = sqrt(2) * r_out

        # inner tube radius is set to r_out_tube - t_p
        self.t_pipe_wall_actual = r_out - r_in
        self.r_in_tube = self.r_out_tube - self.t_pipe_wall_actual

        # r_convection is set to r_in_tube - 1/4 * t_p
        self.r_convection = self.r_in_tube - self.t_pipe_wall_actual / 4.0

        # r_fluid is set to r_in_convection - 3/4 * t_p
        self.r_fluid = self.r_convection - (3.0 / 4.0 * self.t_pipe_wall_actual)

        # Thicknesses of the grid regions
        self.thickness_soil_cell = (self.r_far_field - self.r_borehole) / self.num_soil_cells
        self.thickness_grout_cell = (self.r_borehole - self.r_out_tube) / self.num_grout_cells
        self.thickness_pipe_cell = (self.r_out_tube - self.r_in_tube) / self.num_pipe_cells
        self.thickness_conv_cell = (self.r_in_tube - self.r_convection) / self.num_conv_cells
        self.thickness_fluid_cell = (self.r_convection - self.r_fluid) / self.num_fluid_cells

        # other
        self.init_temp = 20

        # other
        self.g = np.array([], dtype=np.double)
        self.g_bhw = np.array([], dtype=np.double)
        self.lntts = np.array([], dtype=np.double)
        self.c_0 = TWO_PI * self.soil.k
        self.t_s = self.borehole.H**2 / (9 * self.soil.alpha)
        # default is at least 49 hours, or up to -8.6 log time
        self.calc_time_in_sec = max([self.t_s * exp(-8.6), 49.0 * SEC_IN_HR])
        self.g_sts = None

    def calc_fluid_pipe_resistance(self) -> float:
        self.R_fp = self.bhr_borehole.calc_fluid_pipe_resist(self.m_flow_borehole, self.soil.ugt)
        return self.R_fp

    def calc_effective_borehole_resistance(self) -> float:
        resist_bh_effective = self.bhr_borehole.calc_bh_resist(self.m_flow_borehole, self.soil.ugt)
        return resist_bh_effective

    def to_single(self):
        return self

    def as_dict(self) -> dict:
        return {"type": str(self.__class__)}

    def fill_radial_cells(self, resist_f_effective, resist_pg_effective):
        radial_cells = np.zeros(shape=(len(CellProps), self.num_cells), dtype=np.double)

        cell_summation = 0

        def fill_single_cell(inner_radius, thickness, conductivity, rho_cp):
            center_radius = inner_radius + thickness / 2.0
            outer_radius = inner_radius + thickness
            volume = pi * (outer_radius**2 - inner_radius**2)
            return np.array(
                [inner_radius, center_radius, outer_radius, conductivity, rho_cp, self.init_temp, volume],
                dtype=np.double,
            )

        # load fluid cells
        # The equivalent thermal mass of the fluid can be calculated from
        # equation (2)
        # pi (r_in_conv ** 2 - r_f **2) C_eq_f = 2pi r_p_in**2 * C_f
        rho_cp_eq_fluid = 2.0 * (self.pipe.r_in**2) * self.fluid.rho_cp
        rho_cp_eq_fluid /= (self.r_convection**2) - (self.r_fluid**2)
        conductivity_fluid = 200
        for idx in range(cell_summation, self.num_fluid_cells + cell_summation):
            inner_radius_fluid_cell = self.r_fluid + idx * self.thickness_fluid_cell
            radial_cells[:, idx] = fill_single_cell(
                inner_radius_fluid_cell, self.thickness_fluid_cell, conductivity_fluid, rho_cp_eq_fluid
            )

        cell_summation += self.num_fluid_cells

        # load convection cells
        conductivity_conv = log(self.r_in_tube / self.r_convection) / (TWO_PI * resist_f_effective)
        rho_cp_conv = 1.0
        for j, idx in enumerate(range(cell_summation, self.num_conv_cells + cell_summation)):
            inner_radius_conv_cell = self.r_convection + j * self.thickness_conv_cell
            radial_cells[:, idx] = fill_single_cell(
                inner_radius_conv_cell, self.thickness_conv_cell, conductivity_conv, rho_cp_conv
            )

        cell_summation += self.num_conv_cells

        # load pipe cells
        conductivity_pipe_grout = log(self.r_borehole / self.r_in_tube) / (TWO_PI * resist_pg_effective)
        rho_cp_pipe = self.pipe.rho_cp
        for j, idx in enumerate(range(cell_summation, self.num_pipe_cells + cell_summation)):
            inner_radius_pipe_cell = self.r_in_tube + j * self.thickness_pipe_cell
            radial_cells[:, idx] = fill_single_cell(
                inner_radius_pipe_cell, self.thickness_pipe_cell, conductivity_pipe_grout, rho_cp_pipe
            )

        cell_summation += self.num_pipe_cells

        # load grout cells
        rho_cp_grout = self.grout.rho_cp
        for j, idx in enumerate(range(cell_summation, self.num_grout_cells + cell_summation)):
            inner_radius_grout_cell = self.r_out_tube + j * self.thickness_grout_cell
            radial_cells[:, idx] = fill_single_cell(
                inner_radius_grout_cell, self.thickness_grout_cell, conductivity_pipe_grout, rho_cp_grout
            )

        cell_summation += self.num_grout_cells

        # load soil cells
        conductivity_soil = self.soil.k
        rho_cp_soil = self.soil.rho_cp
        for j, idx in enumerate(range(cell_summation, self.num_soil_cells + cell_summation)):
            inner_radius_soil_cell = self.r_borehole + j * self.thickness_soil_cell
            radial_cells[:, idx] = fill_single_cell(
                inner_radius_soil_cell, self.thickness_soil_cell, conductivity_soil, rho_cp_soil
            )

        cell_summation += self.num_soil_cells

        return radial_cells

    @staticmethod
    def solve_tridiagonal(dl, d, du, b, overwrite_b=False):
        n = len(d)
        # Copy or alias b depending on overwrite_b
        dc = b if overwrite_b else np.copy(b)

        # Make copies of diagonals (always safe to copy these)
        ac = np.copy(du)
        bc = np.copy(d)
        cc = np.copy(dl)

        # Forward sweep
        for i in range(1, n):
            w = cc[i - 1] / bc[i - 1]
            bc[i] -= w * ac[i - 1]
            dc[i] -= w * dc[i - 1]

        # Back substitution
        dc[-1] = dc[-1] / bc[-1]
        for i in range(n - 2, -1, -1):
            dc[i] = (dc[i] - ac[i] * dc[i + 1]) / bc[i]

        return dc

    def calc_sts_g_functions(self, final_time=None) -> tuple:
        self.t_s = self.borehole.H**2 / (9 * self.soil.alpha)
        self.calc_time_in_sec = max([self.t_s * exp(-8.6), 49.0 * SEC_IN_HR])

        # effective borehole resistance
        resist_bh_effective = self.calc_effective_borehole_resistance()

        # effective convection resistance, assumes 2 pipes
        fluid_resist = self.bhr_borehole.calc_fluid_resist(self.m_flow_borehole, self.soil.ugt)
        resist_f_effective = fluid_resist / 2.0

        # effective combined pipe-grout resistance. assumes Rees 2016, eq. 3.6 applies
        resist_pg_effective = resist_bh_effective - resist_f_effective

        radial_cells = self.fill_radial_cells(resist_f_effective, resist_pg_effective)

        if final_time is None:
            final_time = self.calc_time_in_sec

        g = []
        g_bhw = []
        lntts = []

        _dl = np.zeros(self.num_cells - 1)
        _d = np.zeros(self.num_cells)
        _du = np.zeros(self.num_cells - 1)
        _b = np.zeros(self.num_cells)

        heat_flux = 1.0
        init_temp = self.init_temp

        time = 1e-12 - 120
        time_step = 120

        _fe_1 = np.zeros(shape=(self.num_cells - 2), dtype=np.double)
        _fe_2 = np.zeros_like(_fe_1)
        _ae = np.zeros_like(_fe_2)
        _fw_1 = np.zeros_like(_ae)
        _fw_2 = np.zeros_like(_fw_1)
        _aw = np.zeros_like(_fw_2)
        _ad = np.zeros_like(_aw)

        _west_cell = radial_cells[:, 0 : self.num_cells - 2]
        _center_cell = radial_cells[:, 1 : self.num_cells - 1]
        _east_cell = radial_cells[:, 2 : self.num_cells - 0]

        fe_1 = log(radial_cells[CellProps.R_OUT, 0] / radial_cells[CellProps.R_CENTER, 0])
        fe_1 /= TWO_PI * radial_cells[CellProps.K, 0]

        fe_2 = log(radial_cells[CellProps.R_CENTER, 1] / radial_cells[CellProps.R_IN, 1])
        fe_2 /= TWO_PI * radial_cells[CellProps.K, 1]

        ae = 1 / (fe_1 + fe_2)
        ad = radial_cells[CellProps.RHO_CP, 0] * radial_cells[CellProps.VOL, 0] / time_step
        _d[0] = -ae / ad - 1
        _du[0] = ae / ad

        def fill_f1(fx_1, cell):
            fx_1[:] = np.log(cell[CellProps.R_OUT, :] / cell[CellProps.R_CENTER, :]) / (TWO_PI * cell[CellProps.K, :])

        def fill_f2(fx_2, cell):
            fx_2[:] = np.log(cell[CellProps.R_CENTER, :] / cell[CellProps.R_IN, :]) / (TWO_PI * cell[CellProps.K, :])

        fill_f1(_fe_1, _center_cell)
        fill_f2(_fe_2, _east_cell)
        _ae[:] = 1.0 / (_fe_1 + _fe_2)

        fill_f1(_fw_1, _west_cell)
        fill_f2(_fw_2, _center_cell)
        _aw[:] = -1.0 / (_fw_1 + _fw_2)

        _ad[:] = _center_cell[CellProps.RHO_CP, :] * _center_cell[CellProps.VOL, :] / time_step
        _dl[0 : self.num_cells - 2] = -_aw / _ad
        _d[1 : self.num_cells - 1] = _aw / _ad - _ae / _ad - 1.0
        _du[1 : self.num_cells - 1] = _ae / _ad

        while True:
            time += time_step

            # For the idx == 0 case:

            _b[0] = -radial_cells[CellProps.TEMP, 0] - heat_flux / ad

            # For the idx == n-1 case

            _dl[self.num_cells - 2] = 0.0
            _d[self.num_cells - 1] = 1.0
            _b[self.num_cells - 1] = radial_cells[CellProps.TEMP, self.num_cells - 1]

            # Now handle the 1 to n-2 cases with numpy slicing and vectorization
            _b[1 : self.num_cells - 1] = -radial_cells[CellProps.TEMP, 1 : self.num_cells - 1]

            # Tri-diagonal matrix solver
            # High level interface to LAPACK routine
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.lapack.dgtsv.html#scipy.linalg.lapack.dgtsv
            dgtsv(_dl, _d, _du, _b, overwrite_b=1)  # TODO: Do we really need lapack just to do a TDMA solution?
            # self.solve_tridiagonal(_dl, _d, _du, _b, overwrite_b=True)

            radial_cells[CellProps.TEMP, :] = _b

            # compute standard g-functions
            g.append(self.c_0 * ((radial_cells[CellProps.TEMP, 0] - init_temp) / heat_flux - resist_bh_effective))

            # compute g-functions at bh wall
            bh_wall_temp = radial_cells[CellProps.TEMP, self.bh_wall_idx]
            g_bhw.append(self.c_0 * ((bh_wall_temp - init_temp) / heat_flux))

            lntts.append(log(time / self.t_s))

            if time >= final_time - time_step:
                break

        # quickly chop down the total values to a more manageable set
        num_intervals = 30
        g_tmp = interp1d(lntts, g)
        uniform_lntts_vals = np.linspace(lntts[0], lntts[-1], num_intervals)
        uniform_g_vals = g_tmp(uniform_lntts_vals)

        g_bhw_tmp = interp1d(lntts, g_bhw)
        uniform_g_bhw_vals = g_bhw_tmp(uniform_lntts_vals)

        # set the final arrays and interpolator objects
        self.lntts = np.array(uniform_lntts_vals)
        self.g = np.array(uniform_g_vals)
        self.g_bhw = np.array(uniform_g_bhw_vals)
        self.g_sts = interp1d(self.lntts, self.g)

        return self.lntts, self.g



#!/usr/bin/env python
import logging
import sys
from json import loads
from pathlib import Path

import click
from jsonschema.exceptions import ValidationError

from ghedesigner.constants import VERSION
from ghedesigner.enums import TimestepType
from ghedesigner.ghe.manager import GroundHeatExchanger
from ghedesigner.heat_pump import HeatPump
from ghedesigner.output.manager import OutputManager
from ghedesigner.utilities import write_idf
from ghedesigner.validate import validate_input_file

logging.basicConfig(level=logging.WARN, format="%(message)s", datefmt="[%X]")
logger = logging.getLogger(__name__)


def run(input_file_path: Path, output_directory: Path) -> int:
    """
    Worker function to run simulation.

    :param input_file_path: path to input file. Input file must exist.
    :param output_directory: path to write output files. Output directory must be a valid path.
    """

    # validate inputs against schema before doing anything
    try:
        validate_input_file(input_file_path)
    except ValidationError:
        return 1

    # read all inputs into a dict
    full_inputs = loads(input_file_path.read_text())

    # Read in all the inputs into small dicts
    # it is possible to define multiple fluids, GHEs, and boreholes in the inputs, I'm just taking the first for now
    input_file_version: int = full_inputs["version"]
    if input_file_version != 2:  # noqa: PLR2004
        print("Bad input file version, right now we support these versions: 1")
        return 1

    # Validate the load source, it should be a building object or a GHE with loads specified
    # any GHE instances found with pre_designed will just be ignored since they don't need anything added
    unsized_ghe_contains_loads = []
    for _, ghe_dict in full_inputs["ground_heat_exchanger"].items():
        if "pre_designed" in ghe_dict:
            continue  # no need for loads checks here, don't even add them to the contains_loads list
        if "loads" in ghe_dict:
            unsized_ghe_contains_loads.append(True)
    all_ghe_has_loads = all(unsized_ghe_contains_loads)
    no_ghe_has_loads = not any(unsized_ghe_contains_loads)
    building_input = "building" in full_inputs
    valid_load_source = all_ghe_has_loads ^ (building_input and no_ghe_has_loads)  # XOR because we don't want both
    if not valid_load_source:
        logger.warning("Bad load specified, need exactly one of: loads in each ghe, or building object")

    # Loop over the topology and init the found objects, for now just the GHE or a GHE with an HP
    topology_props: list[dict] = full_inputs["topology"]
    ghe_names = []
    building_names = []
    for component in topology_props:
        if component["type"] == "building":
            building_names.append(component["name"])
        elif component["type"] == "ground_heat_exchanger":
            ghe_names.append(component["name"])

    # TODO: check on simulation_control, it's not required by schema because it may not be needed

    # do actions depending on what is provided in input
    if len(ghe_names) >= 1 and len(building_names) == 0:
        # we are just doing a GHE design/sizing/simulation alone
        for ghe_name in ghe_names:
            ghe_dict = full_inputs["ground_heat_exchanger"][ghe_name]

            if "loads" in ghe_dict and "file_path" in ghe_dict["loads"]:
                if Path(ghe_dict["loads"]["file_path"]).is_absolute():
                    ghe_dict["loads"]["file_path"] = str(Path(ghe_dict["loads"]["file_path"]).resolve())
                else:
                    # relatives paths as referenced from input file
                    input_file_dir = input_file_path.parent.resolve()
                    relative_file_path = Path(ghe_dict["loads"]["file_path"])
                    loads_path = input_file_dir / relative_file_path
                    ghe_dict["loads"]["file_path"] = str(loads_path.resolve())

            ghe = GroundHeatExchanger.init_from_dictionary(ghe_dict, full_inputs["fluid"])
            if "pre_designed" in ghe_dict:
                log_time, g_values, g_bhw_values = ghe.get_g_function(ghe_dict)
                results = OutputManager("GHEDesigner Run from CLI", "Just Calculate G", "", "")
                results.just_write_g_function(output_directory, log_time, g_values, g_bhw_values)
            else:
                # TODO: Assert that "design" data is in the ghe object
                search, search_time, _ = ghe.design_and_size_ghe(
                    ghe_dict, full_inputs["simulation_control"]["sizing_months"]
                )
                results = OutputManager("GHEDesigner Run from CLI", "Notes", "Author", "Iteration Name")
                results.set_design_data(search, search_time, load_method=TimestepType.HYBRID)
                results.write_all_output_files(output_directory=output_directory, file_suffix="")
    elif len(ghe_names) == 1 and len(building_names) == 1:
        # we have a GHE and a building, grab both
        ghe_dict = full_inputs["ground_heat_exchanger"][ghe_names[0]]
        ghe = GroundHeatExchanger.init_from_dictionary(ghe_dict, full_inputs["fluid"])
        single_building = full_inputs["building"][building_names[0]]
        heat_pump = HeatPump(single_building["name"])
        heat_pump.set_fixed_cop(single_building["cop"])
        loads_file_path = Path(single_building["loads"]).resolve()
        if not loads_file_path.exists():  # TODO: I'll try to find it relative to repo/tests/ for now...
            this_file = Path(__file__).resolve()
            ghe_designer_dir = this_file.parent
            tests_dir = ghe_designer_dir / "tests"
            loads_file_path = tests_dir / single_building["loads"]
        heat_pump.set_loads_from_file(loads_file_path)
        ghe_loads = heat_pump.get_ground_loads()
        if "pre_designed" in ghe_dict:
            log_time, g_values, g_bhw_values = ghe.get_g_function(ghe_dict)
            print(g_values, g_bhw_values)
        else:
            search, search_time, _ = ghe.design_and_size_ghe(
                ghe_dict, full_inputs["simulation_control"]["sizing_months"], loads_override=ghe_loads
            )
            results = OutputManager("GHEDesigner Run from CLI", "Notes", "Author", "Iteration Name")
            results.set_design_data(search, search_time, load_method=TimestepType.HYBRID)
            results.write_all_output_files(output_directory=output_directory, file_suffix="")
    else:
        print("Bad input file, for now the only available configs are: 1 GLHE alone, or 1 GLHE and 1 building")
        return 1
    return 0


@click.command(name="GHEDesignerCommandLine")
@click.argument("input-path", type=click.Path(exists=True), required=True)
@click.argument("output-directory", type=click.Path(exists=False), required=False)
@click.version_option(VERSION)
@click.option("--validate-only", default=False, is_flag=True, show_default=False, help="Validate input file and exit.")
@click.option("-c", "--convert", help="Convert output to specified format. Options supported: 'IDF'.")
def run_manager_from_cli(input_path, output_directory, validate_only, convert):
    # Note that since this is wrapped in click, it should use the exit(code) instead of return.
    # Click will absorb the return code and not return it.
    # If we use exit(code), it will return the code properly.
    input_path = Path(input_path).resolve()

    if validate_only:
        try:
            validate_input_file(input_path)
            logger.info("Valid input file.")
            sys.exit(0)
        except ValidationError as ve:
            logger.error(ve)
            sys.exit(1)

    if convert:
        if convert == "IDF":
            try:
                write_idf(input_path)
                print("Output converted to IDF objects.")
                sys.exit(0)
            except Exception as e:  # noqa: BLE001
                logger.warning(f"Conversion to IDF error: {e}")
                sys.exit(1)

        else:
            print(f"Unsupported conversion format type: {format}", file=sys.stderr)
            sys.exit(1)

    if output_directory is None:
        print("Output directory path must be passed as an argument, aborting", file=sys.stderr)
        sys.exit(1)

    output_path = Path(output_directory).resolve()

    return_code = run(input_path, output_path)
    sys.exit(return_code)


if __name__ == "__main__":
    exit_code = run_manager_from_cli()
    sys.exit(exit_code)
